/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

void set_matrix(MATRIX m, FLOAT_T **m0)
{
  int i,n0=-1,n1=-1;

#ifdef CHECK_DIM
  if ((n0 = dim0(m)) <= 0 || (n1 = dim1(m)) <= 0) DIE_DIM;
#else
  n0 = dim0(m);
  n1 = dim1(m);
#endif

  for (i=0; i<n0; i++)
    memcpy(m[i],m0[i],sizeof(FLOAT_T)*n1);
}

void set_vector(VECTOR v, FLOAT_T *v0)
{
  int n;

#ifdef CHECK_DIM
  if ((n = dim(v)) <= 0) DIE_DIM;
#else
  n = dim(v);
#endif

  memcpy(v,v0,sizeof(FLOAT_T)*n);
}

MATRIX alloc_set_matrix(int n0, int n1, FLOAT_T **m0)
{
  MATRIX m;

#ifdef CHECK_DIM
  if (n0 <= 0 || n1 <= 0) DIE_DIM;
#endif

  m = alloc_matrix(n0,n1);
  set_matrix(m,m0);

  return m;
}

VECTOR alloc_set_vector(int n, FLOAT_T *v0)
{
  VECTOR v;

#ifdef CHECK_DIM
  if (n <= 0) DIE_DIM;
#endif

  v = alloc_vector(n);
  set_vector(v,v0);

  return v;
}

void copy_matrix(MATRIX m_dst, MATRIX m_src)
{
  int i,n0=-1,n1=-1;

#ifdef CHECK_DIM
  if ((n0 = dim0(m_dst)) <= 0 || (n1 = dim1(m_dst)) <= 0 ||
      n0 != dim0(m_src) || n1 != dim1(m_src))
    DIE_DIM;
#else
  n0 = dim0(m_dst);
  n1 = dim1(m_dst);
#endif

  for (i=0; i<n0; i++)
    memcpy(m_dst[i],m_src[i],sizeof(FLOAT_T)*n1);
}

void copy_vector(VECTOR v_dst, VECTOR v_src)
{
  int n;

#ifdef CHECK_DIM
  if ((n = dim(v_dst)) <= 0 || n != dim(v_src)) DIE_DIM;
#else
  n = dim(v_dst);
#endif

  memcpy(v_dst,v_src,sizeof(FLOAT_T)*n);
}

VECTOR dup_vector(VECTOR v)
{
  int n;
  VECTOR new_v;

#ifdef CHECK_DIM
  if ((n = dim(v)) <= 0) DIE_DIM;
#else
  n = dim(v);
#endif

  new_v = alloc_vector(n);
  copy_vector(new_v,v);

  return new_v;
}

MATRIX dup_matrix(MATRIX m)
{
  int n0=-1,n1=-1;
  MATRIX new_m;

#ifdef CHECK_DIM
  if ((n0 = dim0(m)) <= 0 || (n1 = dim1(m)) <= 0) DIE_DIM;
#else
  n0 = dim0(m);
  n1 = dim1(m);
#endif

  new_m = alloc_matrix(n0,n1);
  copy_matrix(new_m,m);

  return new_m;
}

void set_vprod_matrix(MATRIX m, VECTOR v0, VECTOR v1)
{
  int n0=-1,n1=-1,i,j;

#ifdef CHECK_DIM
  if ((n0 = dim(v0)) <= 0 || (n1 = dim(v1)) <= 0 ||
      n0 != dim0(m) || n1 != dim1(m))
    DIE_DIM;
#else
  n0 = dim(v0);
  n1 = dim(v1);
#endif

  for (i=0; i<n0; i++)
    for (j=0; j<n1; j++)
      m[i][j] = v0[i] * v1[j];
}

MATRIX alloc_vprod_matrix(VECTOR v0, VECTOR v1)
{
  int n0=-1,n1=-1;
  MATRIX m;

#ifdef CHECK_DIM
  if ((n0 = dim(v0)) <= 0 || (n1 = dim(v1)) <= 0) DIE_DIM;
#else
  n0 = dim(v0);
  n1 = dim(v1);
#endif

  m = alloc_matrix(n0,n1);
  set_vprod_matrix(m,v0,v1);

  return m;
}

VECTOR alloc_proj_vector(VECTOR v, int *use)
{
  return alloc_proj_vector_n(v,-1,use);
}

VECTOR alloc_proj_vector_n(VECTOR v, int n_use, int *use)
{
  int i,n=-1;
  VECTOR v_new;

#ifdef CHECK_DIM
  if ((n = dim(v)) <= 0 || n_use > n) DIE_DIM;
#else
  n = dim(v);
#endif

  if (n_use < 0) {
    n_use = 0;
    for (i=0; i<n; i++) if (use[i]) n_use++;
  }

  v_new = alloc_vector(n_use);
  proj_vector(v,use,v_new);

  return v_new;
}

MATRIX alloc_proj_matrix(MATRIX m, int *use0, int *use1)
{
  return alloc_proj_matrix_n(m,-1,-1,use0,use1);
}

MATRIX
alloc_proj_matrix_n(MATRIX m, int n_use0, int n_use1, int *use0, int *use1)
{
  int i=0,j=0,n0=-1,n1=-1;
  MATRIX m_new;

#ifdef CHECK_DIM
  if ((n0 = dim0(m)) <= 0 || (n1 = dim(m)) <= 0 || n_use0 > n0 || n_use1 > n1)
    DIE_DIM;
#else
  n0 = dim0(m);
  n1 = dim1(m);
#endif

  if (n_use0 < 0) {
    n_use0 = 0;
    for (i=0; i<n0; i++) if (use0[i]) n_use0++;
  }
  if (n_use1 < 0) {
    n_use1 = 0;
    for (j=0; j<n1; j++) if (use1[i]) n_use1++;
  }

  m_new = alloc_matrix(n_use0,n_use1);
  proj_matrix(m,use0,use1,m_new);

  return m_new;
}

void clear_matrix(MATRIX m)
{
  int n0=-1,n1=-1;

#ifdef CHECK_DIM
  if ((n0 = dim0(m)) <= 0 || (n1 = dim1(m)) <= 0) DIE_DIM;
#else
  n0 = dim0(m);
  n1 = dim1(m);
#endif

  memset(m[0],0,sizeof(FLOAT_T)*n0*n1);
}

void clear_vector(VECTOR v)
{
  int n;

#ifdef CHECK_DIM
  if ((n = dim(v)) <= 0) DIE_DIM;
#else
  n = dim(v);
#endif

  memset(v,0,sizeof(FLOAT_T)*n);
}

VECTOR alloc_clear_vector(int n)
{
  VECTOR v;

#ifdef CHECK_DIM
  if (n <= 0) DIE_DIM;
#endif

  v = alloc_vector(n);
  clear_vector(v);

  return v;
}

MATRIX alloc_clear_matrix(int n0, int n1)
{
  MATRIX m;

#ifdef CHECK_DIM
  if (n0 <= 0 || n1 <= 0) DIE_DIM;
#endif

  m = alloc_matrix(n0,n1);
  clear_matrix(m);

  return m;
}

#ifdef DISABLE_DIM

int fdim0(MATRIX m)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  M_HEADER *b;

  if (m == NULL) return 0;

  b = (M_HEADER *)m - 1;

#ifdef CHECK_DIM
  if (b < matrix_pool || b > mp_last) {
    fprintf(stderr,"Internal error: dim0()"
            " -- illegal address (%p) of a matrix (pool = %p, last = %p)\n",
            b,vector_pool,vp_last);
    exit(1);
  }
#endif

  return b->info.dim0;
}

int fdim1(MATRIX m)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  M_HEADER *b;

  if (m == NULL) return 0;

  b = (M_HEADER *)m - 1;

#ifdef CHECK_DIM
  if (b < matrix_pool || b > mp_last) {
    fprintf(stderr,"Internal error: dim1()"
            " -- illegal address (%p) of a matrix (pool = %p, last = %p)\n",
            b,vector_pool,vp_last);
    exit(1);
  }
#endif

  return b->info.dim1;
}

int fdim(VECTOR v)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  V_HEADER *b;

  if (v == NULL) return 0;

  b = (V_HEADER *)v - 1;

#ifdef CHECK_DIM
  if (b < vector_pool || b > vp_last) {
    fprintf(stderr,
            "Internal error: dim() -- illegal address (%p) of a vector"
            " (vector_pool = %p, vp_last = %p)\n",
            b,vector_pool,vp_last);
    exit(1);
  }
#endif

  return b->info.dim;
}

#endif /* DISABLE_DIM */
