/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

int is_decreased(int r, int m, int n)
{
  /* This would not happen without a bug or a numeric error */
  if (mode_daem &&
      log_itemp[m] - log_itemp[m-1] < ACCEPT_SMALL_NUMBER &&
      lambda[m] < lambda[m-1] &&
      (lambda[m-1] - lambda[m]) > ACCEPT_SMALL_NUMBER) {
    print_decrease(fp_log,r,m,lambda[m-1],lambda[m],log_itemp[m]);
    print_param(fp_log);
    die(conv_msg(" (iter:%d) LL or LPost decreased! "
                 "(before: %.Nf (%.Nf) | after: %.Nf (%.Nf) |"
                 " diff: %.Nf) -- see the log file\n"),
        m,lambda[m-1],log_itemp[m-1],
        lambda[m],log_itemp[m],(lambda[m-1] - lambda[m]));
    return 1;
  }
  else if (!mode_daem &&
           lambda[m] < lambda[m-1] &&
           (lambda[m-1] - lambda[m]) > ACCEPT_SMALL_NUMBER) {
    print_decrease(fp_log,r,m,lambda[m-1],lambda[m],0.0);
    print_param(fp_log);
    die(conv_msg(" (iter:%d) LL or LPost decreased! "
                 "(before: %.Nf | after: %.Nf | diff: %.Nf)"
                 " -- see the log file\n"),
        m,lambda[m-1],lambda[m],(lambda[m-1] - lambda[m]));
    return 1;
  }
  return 0;
}

int is_converged(int r, int m, int n)
{
  if (m >= n) {
    print_msg(" (Stopped: %d iters)\tL=%.Nf\n",m,lambda[m]);
    return 1;
  }
  /* [Note]
   *   We also stop EM iterations if the log-likelihood (or the a posteriori
   *   probability) decreased.  Of course, the magnitude of decrease is not
   *   negligible, it should have been detected by is_decreased() in advance.
   */
  else if (lambda[m] - lambda[m-1] < epsilon) {
    print_msg(" (Converged: %d iters)\tL=%.Nf\n",m,lambda[m]);
    return 1;
  }
  return 0;
}

/* m: total number of iterations,
 * mm: number of iterations at the current inverse temperature
 * itemp: should be =< 1.0
 */
int is_converged_daem(int r, int m, int mm, int n, FLOAT_T itemp)
{
  if (itemp < 1.0) {
    if (mm >= n) {
      return 1;
    }
    else if (log_itemp[m] - log_itemp[m-1] < ACCEPT_SMALL_NUMBER &&
             lambda[m] - lambda[m-1] > 0 &&
             lambda[m] - lambda[m-1] < epsilon) {
      return 1;
    }
  }
  else {
    if (mm >= n) {
      print_msg(" (Stopped: %d iters)\tL=%.Nf\n",m,lambda[m]);
      return 1;
    }
    else if (log_itemp[m] - log_itemp[m-1] < ACCEPT_SMALL_NUMBER &&
             lambda[m] - lambda[m-1] < epsilon &&
             lambda[m] - lambda[m-1] > 0) {
      print_msg(" (Converged: %d iters)\tL=%.Nf\n",m,lambda[m]);
      return 1;
    }
  }

  return 0;
}

/* returns the existence of singularity (including collinearity) */
static int is_singular(int r, int m, int n)
{
  int k,u;

  for (k=0; k<num_class; k++) {
    for (u=0; u<gauss_size; u++) {
      if (gauss_attr_size[u] > 1 &&
          FABS(param_sdet[k][u]) < VERY_SMALL_FLOAT) {
        print_msg(" (Stopped: %d iters -- singular covars (det=%.Ne) "
                  "at class#=%d, gauss#=%d)\tL=%.Nf\n",
                  m,param_sdet[k][u],k,u,lambda[m]);
        return 1;
      }
    }
  }

  return 0;
}

#if 0
static void check_param(void)
{
  int k,v,i;
  FLOAT_T sum_c, sum_w;

  sum_c = 0.0;
  for (k=0; k<num_class; k++)
    sum_c += param_c[k];

  print_msg("check> sum_c = %.Nf\n",sum_c);

  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size_d; v++) {
      sum_w = 0.0;
      for (i=0; i<num_word[v]; i++)
        sum_w += param_w[k][v][i];
      print_msg("check> sum_w1[%d] = %.Nf\n",k,sum_w);
    }
  }
}
#endif

static void save_loglike_best(int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++) {
    loglike_best[m] = loglike[m];
    lambda_best[m] = lambda[m];
  }

  if (mode_smooth)
    for (m=0; m<num_iterate; m++)
      log_prior_best[m] = log_prior[m];
}

static void restore_loglike_best(int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++) {
    loglike[m] = loglike_best[m];
    lambda[m] = lambda_best[m];
  }

  if (mode_smooth)
    for (m=0; m<num_iterate; m++)
      log_prior[m] = log_prior_best[m];
}

void save_log_itemp_best(int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++)
    log_itemp_best[m] = log_itemp[m];
}

void restore_log_itemp_best(int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++)
    log_itemp[m] = log_itemp_best[m];
}

static void compute_param_precisions(void)
{
  int k,u;

#pragma omp parallel for private(u)
  for (k=0; k<num_class; k++) {
    for (u=0; u<gauss_size; u++) {
      param_sdet[k][u] = INVERT(param_s[k][u],param_p[k][u]);
      if (err_inv) die("Failed to invert covariances\n");
    }
  }
}

void initialize_param(void)
{
  int k,v,i,u,a,n;
  FLOAT_T r,sum,noise;

  if (init_type_class == RANDOM_C) {
    sum = 0.0;
    for (k=0; k<num_class; k++) {
      r = random_int() + RAND_MAX;
      param_c[k] = r;
      sum += r;
    }
    for (k=0; k<num_class; k++)
      param_c[k] /= sum;
  }
  else if (init_type_class == NOISY_UNIFORM_C) {
    sum = 0.0;
    for (k=0; k<num_class; k++) {
      noise = noise_ratio * random_float();
      r = (1.0 / num_class) * (1 + noise);
      param_c[k] = r;
      sum += r;
    }
    for (k=0; k<num_class; k++)
      param_c[k] /= sum;
  }

  if (mode_nominal) {
    if (init_type_word == RANDOM_W) {
      for (v=0; v<tuple_size_d; v++) {
        for (k=0; k<num_class; k++) {
          sum = 0.0;
          for (i=0; i<num_word[v]; i++) {
            r = random_int() + RAND_MAX;
            param_w[k][v][i] = r;
            sum += r;
          }
          for (i=0; i<num_word[v]; i++)
            param_w[k][v][i] /= sum;
        }
      }
    }
    else if (init_type_word == NOISY_UNIFORM_W) {
      for (v=0; v<tuple_size_d; v++) {
        for (k=0; k<num_class; k++) {
          sum = 0.0;
          for (i=0; i<num_word[v]; i++) {
            noise = noise_ratio * random_float();
            r = (1.0 / num_word[v]) * (1 + noise);
            param_w[k][v][i] = r;
            sum += r;
          }
          for (i=0; i<num_word[v]; i++)
            param_w[k][v][i] /= sum;
        }
      }
    }
    else if (init_type_word == WORD_FREQ) {
      for (v=0; v<tuple_size_d; v++) {
        for (k=0; k<num_class; k++) {
          sum = 0.0;
          for (i=0; i<num_word[v]; i++) {
            noise = noise_ratio * random_float() / num_word[v];
            r = 1.0 / word_freq_list[v][i] + noise;
            param_w[k][v][i] = r;
            sum += r;
          }
          for (i=0; i<num_word[v]; i++)
            param_w[k][v][i] /= sum;
        }
      }
    }
    else
      die("Internal error:"
          " unexpected initialization method for discrete attribute values\n");
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];

        copy_vector(param_u[k][u],sample_u[u]);
        copy_matrix(param_s[k][u],sample_s[u]);

        /* We only add noises to means */
        for (a=0; a<gauss_attr_size[u]; a++) {
          noise =
            noise_ratio_gauss * SQRT(sample_s[u][a][a]) *
            (random_float() - 0.5);
          param_u[k][u][a] += noise;
        }
      }
    }
  }

  if (mode_gauss) compute_param_precisions();

  if (mode_verbose) print_param(fp_log);
}

static void compute_complete_prob_log(void)
{
  int k,e,v,u,n;
  VECTOR x,p_u;
  MATRIX p_s,p_p;
  int *obs,n_miss;
  FLOAT_T det;

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      prob_comp[k][e] = LOG(param_c[k]);

  if (mode_nominal) {
#pragma omp parallel for private(e,v)
    for (k=0; k<num_class; k++) {
      for (e=0; e<num_entry; e++) {
        if (total_num_miss_w > 0 && freq_list[e].num_miss_w > 0) {
          for (v=0; v<tuple_size_d; v++) {
            if (!freq_list[e].miss_w[v])
              prob_comp[k][e] += LOG(param_w[k][v][freq_list[e].id[v]]);
          }
        }
        else {
          for (v=0; v<tuple_size_d; v++)
            prob_comp[k][e] += LOG(param_w[k][v][freq_list[e].id[v]]);
        }
      }
    }
  }

  if (mode_gauss) {
#pragma omp parallel for private(u,n,p_u,p_s,p_p,det,x,e,n_miss,obs)
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];

        p_u = param_u[k][u];
        p_s = param_s[k][u];
        p_p = param_p[k][u];
        det = param_sdet[k][u];

        for (e=0; e<num_entry; e++) {
          x = freq_list[e].gval[u];
          n_miss = freq_list[e].num_miss_g[u];

          if (0 < n_miss && n_miss < n) {
            obs = freq_list[e].obs_g[u];
            prob_comp[k][e] += lgauss_marginal(x,p_u,p_s,obs);
          }
          else if (n_miss == 0) {
            prob_comp[k][e] += lgauss_prec(x,p_u,p_s,p_p,det);
          }
          else {
            /* if n_miss == n, do nothing */
          }
        }
      }
    }
  }

  if (mode_verbose) print_complete_prob(fp_log);
}

static void compute_complete_prob_nolog(void)
{
  int k,e,v,u,n;
  VECTOR x,p_u;
  MATRIX p_s,p_p;
  int *obs,n_miss;
  FLOAT_T det;

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      prob_comp[k][e] = param_c[k];

  if (mode_nominal) {
#pragma omp parallel for private(e,v)
    for (k=0; k<num_class; k++) {
      for (e=0; e<num_entry; e++) {
        if (total_num_miss_w > 0 && freq_list[e].num_miss_w > 0) {
          for (v=0; v<tuple_size_d; v++) {
            if (!freq_list[e].miss_w[v])
              prob_comp[k][e] *= param_w[k][v][freq_list[e].id[v]];
          }
        }
        else {
          for (v=0; v<tuple_size_d; v++)
            prob_comp[k][e] *= param_w[k][v][freq_list[e].id[v]];
        }
      }
    }
  }

  if (mode_gauss) {
#pragma omp parallel for private(u,n,p_u,p_s,p_p,det,x,e,n_miss,obs)
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];

        p_u = param_u[k][u];
        p_s = param_s[k][u];
        p_p = param_p[k][u];
        det = param_sdet[k][u];

        for (e=0; e<num_entry; e++) {
          x = freq_list[e].gval[u];
          n_miss = freq_list[e].num_miss_g[u];

          if (0 < n_miss && n_miss < n) {
            obs = freq_list[e].obs_g[u];
            prob_comp[k][e] *= gauss_marginal(x,p_u,p_s,obs);
          }
          else if (n_miss == 0) {
            prob_comp[k][e] *= gauss_prec(x,p_u,p_s,p_p,det);
          }
          else {
            /* if n_miss == n, do nothing */
          }
        }
      }
    }
  }

  if (mode_verbose) print_complete_prob(fp_log);
}

void compute_complete_prob(void)
{
  if (mode_log_scale && mode_smooth_strict)
    compute_complete_prob_log();
  else
    compute_complete_prob_nolog();
}

static void compute_complete_prob_daem_log(void)
{
  int k,e,v,u,n;
  int n_miss;
  int *obs;
  VECTOR p_u,x;
  MATRIX p_s,p_p;
  FLOAT_T det;

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      prob_comp[k][e] = itemp * LOG(param_c[k]);

  if (mode_nominal) {
#pragma omp parallel for private(e,v)
    for (k=0; k<num_class; k++) {
      for (e=0; e<num_entry; e++) {
        if (total_num_miss_w > 0 && freq_list[e].num_miss_w > 0) {
          for (v=0; v<tuple_size_d; v++) {
            if (!freq_list[e].miss_w[v])
              prob_comp[k][e] += itemp * LOG(param_w[k][v][freq_list[e].id[v]]);
          }
        }
        else {
          for (v=0; v<tuple_size_d; v++)
            prob_comp[k][e] += itemp * LOG(param_w[k][v][freq_list[e].id[v]]);
        }
      }
    }
  }

  if (mode_gauss) {
#pragma omp parallel for private(u,n,p_u,p_s,p_p,det,x,e,n_miss,obs)
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];

        p_u = param_u[k][u];
        p_s = param_s[k][u];
        p_p = param_p[k][u];
        det = param_sdet[k][u];

        for (e=0; e<num_entry; e++) {
          x = freq_list[e].gval[u];
          n_miss = freq_list[e].num_miss_g[u];

          if (0 < n_miss && n_miss < n) {
            obs = freq_list[e].obs_g[u];
            prob_comp[k][e] += itemp * lgauss_marginal(x,p_u,p_s,obs);
          }
          else if (n_miss == 0) {
            prob_comp[k][e] += itemp * lgauss_prec(x,p_u,p_s,p_p,det);
          }
          else {
            /* if n_miss == n, do nothing */
          }
        }
      }
    }
  }

  if (mode_verbose) print_complete_prob(fp_log);
}

static void compute_complete_prob_daem_nolog(void)
{
  int k,e,v,u,n;
  int n_miss;
  int *obs;
  VECTOR p_u,x;
  MATRIX p_s,p_p;
  FLOAT_T det;

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      prob_comp[k][e] = POW(param_c[k],itemp);

  if (mode_nominal) {
#pragma omp parallel for private(e,v)
    for (k=0; k<num_class; k++) {
      for (e=0; e<num_entry; e++) {
        if (total_num_miss_w > 0 && freq_list[e].num_miss_w > 0) {
          for (v=0; v<tuple_size_d; v++) {
            if (!freq_list[e].miss_w[v])
              prob_comp[k][e] *= POW(param_w[k][v][freq_list[e].id[v]],itemp);
          }
        }
        else {
          for (v=0; v<tuple_size_d; v++)
            prob_comp[k][e] *= POW(param_w[k][v][freq_list[e].id[v]],itemp);
        }
      }
    }
  }

  if (mode_gauss) {
#pragma omp parallel for private(u,n,p_u,p_s,p_p,det,x,e,n_miss,obs)
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];

        p_u = param_u[k][u];
        p_s = param_s[k][u];
        p_p = param_p[k][u];
        det = param_sdet[k][u];

        for (e=0; e<num_entry; e++) {
          x = freq_list[e].gval[u];
          n_miss = freq_list[e].num_miss_g[u];

          if (0 < n_miss && n_miss < n) {
            obs = freq_list[e].obs_g[u];
            prob_comp[k][e] *= POW(gauss_marginal(x,p_u,p_s,obs),itemp);
          }
          else if (n_miss == 0) {
            prob_comp[k][e] *= POW(gauss_prec(x,p_u,p_s,p_p,det),itemp);
          }
          else {
            /* if n_miss == n, do nothing */
          }
        }
      }
    }
  }

  if (mode_verbose) print_complete_prob(fp_log);
}

void compute_complete_prob_daem(void)
{
  if (mode_log_scale && mode_smooth_strict)
    compute_complete_prob_daem_log();
  else
    compute_complete_prob_daem_nolog();
}

static void compute_incomplete_prob_log(void)
{
  int k,e;
  FLOAT_T pcomp_max,sum;

  /* OpenMP does not work here XXXX */
  for (e=0; e<num_entry; e++) {
    pcomp_max = 0.0;
    for (k=0; k<num_class; k++)
      if (k == 0 || prob_comp[k][e] > pcomp_max) pcomp_max = prob_comp[k][e];

    sum = 0.0;
    for (k=0; k<num_class; k++)
      sum += EXP(prob_comp[k][e] - pcomp_max);

    prob_incomp[e] = pcomp_max + LOG(sum);
  }

  if (mode_verbose) print_incomplete_prob(fp_log);
}

static void compute_incomplete_prob_nolog(void)
{
  int k,e;

#pragma omp parallel for private(k)
  for (e=0; e<num_entry; e++) {
    prob_incomp[e] = 0.0;
    for (k=0; k<num_class; k++)
      prob_incomp[e] += prob_comp[k][e];
  }

  if (mode_verbose) print_incomplete_prob(fp_log);
}

void compute_incomplete_prob(void)
{
  if (mode_log_scale && mode_smooth_strict)
    compute_incomplete_prob_log();
  else
    compute_incomplete_prob_nolog();
}

FLOAT_T compute_loglike(void)
{
  FLOAT_T ll;
  int e, f;

  ll = 0.0;
  /* doesn't work if a `pragma omp' statement is annotated (i don't know why) */
  if (mode_log_scale && mode_smooth_strict) {
    for (e=0; e<num_entry; e++) {
      f = freq_list[e].freq;
      if (f > 0) ll += f * prob_incomp[e];
    }
  }
  else {
    for (e=0; e<num_entry; e++) {
      f = freq_list[e].freq;
      if (f > 0) ll += f * LOG(prob_incomp[e]);
    }
  }

  if (ISNAN(ll)) {
    print_param(fp_log);
    DIE_NUMERIC("Computed log-likelihood is not a number"
                " -- see the log file\n");
  }
  if (ISINF(ll)) {
    print_param(fp_log);
    DIE_NUMERIC("Computed log-likelihood is not finite -- see the log file\n");
  }

  return ll;
}

FLOAT_T compute_log_prior(void)
{
  FLOAT_T lp,lp_c,lp_w;
  int k,i,v,u,n;
  FLOAT_T alpha_sum,alpha_w0,tau0;
  MATRIX p_s;

  lp = 0.0;

  if (mode_read_hparam) {
    lp_c = 0.0;

    /* Compute the normalizing constant for class */
    alpha_sum = 0.0;
    for (k=0; k<num_class; k++)
      alpha_sum += hparam_c[k];
    lp_c += LGAMMA(alpha_sum);
    for (k=0; k<num_class; k++)
      lp_c -= LGAMMA(hparam_c[k]);

    for (k=0; k<num_class; k++)
      lp_c += (hparam_c[k] - 1.0) * LOG(param_c[k]);

    lp += lp_c;
  }
  else if (smooth_c > 0) {
    lp_c = 0.0;

    lp_c += LGAMMA(num_class * (smooth_c + 1.0));
    lp_c -= num_class * LGAMMA(smooth_c + 1.0);

    for (k=0; k<num_class; k++)
      lp_c += smooth_c * LOG(param_c[k]);

    lp += lp_c;
  }

  lp_w = 0.0;

  if (mode_nominal) {
    if (mode_read_hparam) {

      /* Compute the normalizing constant */
      for (k=0; k<num_class; k++) {
        for (v=0; v<tuple_size_d; v++) {
          alpha_sum = 0.0;
          for (i=0; i<num_word[v]; i++)
            alpha_sum += hparam_w[k][v][i];

          lp_w += LGAMMA(alpha_sum);
        }
      }
      for (k=0; k<num_class; k++)
        for (v=0; v<tuple_size_d; v++)
          for (i=0; i<num_word[v]; i++)
            lp_w -= LGAMMA(hparam_w[k][v][i]);

      for (k=0; k<num_class; k++)
        for (v=0; v<tuple_size_d; v++)
          for (i=0; i<num_word[v]; i++)
            lp_w += (hparam_w[k][v][i] - 1.0) * LOG(param_w[k][v][i]);
    }
    else {

      /* Compute the normalizing constant */
      for (v=0; v<tuple_size_d; v++) {
        lp_w += num_class * LGAMMA(num_word[v] * (smooth_w[v] + 1.0));
        lp_w -= num_class * num_word[v] * LGAMMA(smooth_w[v] + 1.0);
      }

      for (v=0; v<tuple_size_d; v++) {
        if (smooth_w[v] > 0) {
          for (k=0; k<num_class; k++)
            for (i=0; i<num_word[v]; i++)
              lp_w += smooth_w[v] * LOG(param_w[k][v][i]);
        }
      }
    }
  }

  if (mode_gauss && mode_smooth_gauss) {
    /* Compute the Normal-Wishart distribution */
    for (k=0; k<num_class; k++) {
      tau0 = get_tau(k);

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];

        p_s = alloc_set_matrix(n,n,param_s[k][u]); /* covariances */
        mul_const_matrix(p_s,1.0/tau0);

        alpha_w0 = get_alpha_w(k,u,n);

        lp_w += lgauss(param_u[k][u],hparam_u[k][u],p_s);
        lp_w += lwishart_det(param_p[k][u],
                             hparam_s[k][u],hparam_p[k][u],
                             1/param_sdet[k][u],1/hparam_sdet[k][u],
                             alpha_w0);

        release_matrix(p_s);
      }
    }
  }

  lp += lp_w;

  if (ISNAN(lp)) {
    print_param(fp_log);
    DIE_NUMERIC("Computed log-prior is not a number -- see the log file\n");
  }
  if (ISINF(lp)) {
    print_param(fp_log);
    DIE_NUMERIC("Computed log-prior is not finite -- see the log file\n");
  }

  return lp;
}

static void compute_conditional_prob_log(void)
{
  int k,e;

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      prob_cond[k][e] = EXP(prob_comp[k][e] - prob_incomp[e]);

  if (mode_verbose) print_conditional_prob(fp_log);
}

static void compute_conditional_prob_nolog(void)
{
  int k,e;
  FLOAT_T p;

#pragma omp parallel for private(e,p)
  for (k=0; k<num_class; k++) {
    for (e=0; e<num_entry; e++) {
      p = prob_incomp[e];
      prob_cond[k][e] =
        (p > TINY_FLOAT * TINY_FLOAT) ? prob_comp[k][e]/p : 0.0;
    }
  }

  if (mode_verbose) print_conditional_prob(fp_log);
}

void compute_conditional_prob(void)
{
  if (mode_log_scale && mode_smooth_strict)
    compute_conditional_prob_log();
  else
    compute_conditional_prob_nolog();
}

void compute_expect_gauss0(int k, int e, FLOAT_T p_e)
{
  int u,a,a0,a1,a_o,a_m;
  int n,n_obs,n_miss;
  int *miss,*obs;
  VECTOR x,p_u,x_o,x_m,p_u_o,p_u_m,z_o,z_m;
  MATRIX p_s,p_s_oo,p_s_mo,p_s_om,p_s_mm,p_s_oo_inv;
  MATRIX xx,xx_oo,xx_om,xx_mo,xx_mm,z_mo,z_mm,z_mom;
  FLOAT_T v_o,v_m,v_oo,v_om,v_mm;
  FLOAT_T *q_u,**q_s;
  int *rev_ind;
  int max_gauss_attr_size;

  max_gauss_attr_size = -1;
  for (u=0; u<gauss_size; u++) {
    max_gauss_attr_size =
      (u == 0 || gauss_attr_size[u] > max_gauss_attr_size) ?
      gauss_attr_size[u] : max_gauss_attr_size;
  }

  rev_ind = (int *)xmalloc(sizeof(int)*max_gauss_attr_size);

  for (u=0; u<gauss_size; u++) {
    n = gauss_attr_size[u];
    n_miss = freq_list[e].num_miss_g[u];
    n_obs = (n - n_miss);
    x = freq_list[e].gval[u];

    if (n_miss == 0) {      /* all observed */
      for (a=0; a<n; a++)
        expect_u[k][u][a] += p_e * x[a];
      for (a0=0; a0<n; a0++)
        for (a1=a0; a1<n; a1++) /* compute only the upper triangle */
          expect_s[k][u][a0][a1] += p_e * x[a0] * x[a1];
    }
    else if (n == n_miss) { /* all missing */
      for (a=0; a<n; a++)
        expect_u[k][u][a] += p_e * param_u[k][u][a];
      for (a0=0; a0<n; a0++)
        for (a1=a0; a1<n; a1++)
          expect_s[k][u][a0][a1] +=
            p_e *
            (param_u[k][u][a0] * param_u[k][u][a1] + param_s[k][u][a0][a1]);
    }
    else { /* partially observed */
      obs = freq_list[e].obs_g[u];
      miss = freq_list[e].miss_g[u];

      if (n == 2) {
        a_o = (obs[0]) ? 0 : 1;
        a_m = (obs[0]) ? 1 : 0;

        q_u = param_u[k][u];
        q_s = param_s[k][u];

        v_o = x[a_o];
        v_m = (q_u[a_m] + q_s[a_o][a_m] / q_s[a_o][a_o] * (x[a_o] - q_u[a_o]));

        v_oo = x[a_o] * x[a_o];
        v_om = x[a_o] * v_m;
        v_mm =
          v_m * v_m +
          (q_s[a_m][a_m] - q_s[a_o][a_m] * q_s[a_o][a_m] / q_s[a_o][a_o]);

        expect_u[k][u][a_o] += p_e * v_o;
        expect_u[k][u][a_m] += p_e * v_m;
        expect_s[k][u][a_o][a_o] += p_e * v_oo;
        expect_s[k][u][a_m][a_m] += p_e * v_mm;
        expect_s[k][u][0][1]     += p_e * v_om;
      }
      else {
        /* compute the local indices */
        a_o = 0;
        a_m = 0;
        for (a=0; a<n; a++)
          rev_ind[a] = (obs[a]) ? a_o++ : a_m++;

        p_u = param_u[k][u];
        p_s = param_s[k][u];

        x_o = alloc_proj_vector_n(x,n_obs,obs);
        x_m = alloc_vector(n_miss);

        p_u_o = alloc_proj_vector_n(p_u,n_obs,obs);
        p_u_m = alloc_proj_vector_n(p_u,n_miss,miss);

        p_s_oo = alloc_proj_matrix_n(p_s,n_obs,n_obs,obs,obs);
        p_s_mo = alloc_proj_matrix_n(p_s,n_miss,n_obs,miss,obs);
        p_s_om = alloc_proj_matrix_n(p_s,n_obs,n_miss,obs,miss);
        p_s_mm = alloc_proj_matrix_n(p_s,n_miss,n_miss,miss,miss);

        p_s_oo_inv = alloc_matrix(n_obs,n_obs);

        z_o = alloc_vector(n_obs);
        z_m = alloc_vector(n_miss);
        z_mo = alloc_matrix(n_miss,n_obs);
        z_mom = alloc_matrix(n_miss,n_miss);
        z_mm = alloc_matrix(n_miss,n_miss);

        INVERT(p_s_oo,p_s_oo_inv);
        if (err_inv)
          DIE_NUMERIC("Failed to invert covariances of the observed part\n");
        prod_matrix(p_s_mo,p_s_oo_inv,z_mo);

        diff_vector(x_o,p_u_o,z_o);
        prod_matrix_vector(z_mo,z_o,z_m);
        sum_vector(p_u_m,z_m,x_m);

        prod_matrix(z_mo,p_s_om,z_mom);
        diff_matrix(p_s_mm,z_mom,z_mm);

        for (a=0; a<n; a++)
          expect_u[k][u][a] +=
            p_e * ((obs[a]) ? x_o[rev_ind[a]] : x_m[rev_ind[a]]);

        xx_oo = alloc_vprod_matrix(x_o,x_o);
        xx_mo = alloc_vprod_matrix(x_m,x_o);
        xx_om = alloc_vprod_matrix(x_o,x_m);
        xx_mm = alloc_vprod_matrix(x_m,x_m);
        sum_matrix(xx_mm,z_mm,NULL);

        for (a0=0; a0<n; a0++) {
          for (a1=a0; a1<n; a1++) {  /* compute only the upper triangle */
            if (obs[a0] && obs[a1])
              xx = xx_oo;
            else if (obs[a0] && miss[a1])
              xx = xx_om;
            else if (miss[a0] && obs[a1])
              xx = xx_mo;
            else
              xx = xx_mm;

            expect_s[k][u][a0][a1] += p_e * xx[rev_ind[a0]][rev_ind[a1]];
          }
        }

        /* From an implementational reason, it is efficient to release
               matrices/vectors in the same order as that for allocation */

        release_vector(x_o);
        release_vector(x_m);

        release_vector(p_u_o);
        release_vector(p_u_m);
        release_matrix(p_s_oo);
        release_matrix(p_s_mo);
        release_matrix(p_s_om);
        release_matrix(p_s_mm);
        release_matrix(p_s_oo_inv);

        release_vector(z_o);
        release_vector(z_m);
        release_matrix(z_mo);
        release_matrix(z_mom);
        release_matrix(z_mm);

        release_matrix(xx_oo);
        release_matrix(xx_mo);
        release_matrix(xx_om);
        release_matrix(xx_mm);
      }
    }
  }

  xfree(rev_ind);
}

void compute_expect_gauss(void)
{
  int k,e,u;

#pragma omp parallel for private(u)
  for (k=0; k<num_class; k++) {
    for (u=0; u<gauss_size; u++) {
      clear_vector(expect_u[k][u]);
      clear_matrix(expect_s[k][u]);
    }
  }

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      compute_expect_gauss0(k,e,(freq_list[e].freq * prob_cond[k][e]));
}

void compute_expect(void)
{
  int k,i,v,e;
  FLOAT_T tmp_e;

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++) {
    expect_c[k] = 0.0;
    for (e=0; e<num_entry; e++)
      expect_c[k] += freq_list[e].freq * prob_cond[k][e];
  }

  if (mode_nominal) {
#pragma omp parallel for private(i,e,v,tmp_e)
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          expect_w[k][v][i] = 0.0;

      for (e=0; e<num_entry; e++) {
        tmp_e = freq_list[e].freq * prob_cond[k][e];

        if (total_num_miss_w > 0 && freq_list[e].num_miss_w > 0) {
          for (v=0; v<tuple_size_d; v++) {
            if (!freq_list[e].miss_w[v])
              expect_w[k][v][freq_list[e].id[v]] += tmp_e;
          }
        }
        else {
          for (v=0; v<tuple_size_d; v++)
            expect_w[k][v][freq_list[e].id[v]] += tmp_e;
        }
      }
    }
  }

  if (mode_gauss) compute_expect_gauss();

  if (mode_verbose) print_expectation(fp_log);
}

void update_param(void)
{
  FLOAT_T denom_c;
  FLOAT_T denom_w;
  int k,v,i,u,a,a0,a1,n;
  FLOAT_T tmp_s,alpha_w0,hparam_u0,hparam_u1,tau0;
  FLOAT_T e_c_k;
  VECTOR p_u,e_u;
  MATRIX p_s,e_s,h_s;

  if (mode_read_hparam) {
    denom_c = 0.0;
    for (k=0; k<num_class; k++)
      denom_c += (param_c[k] = expect_c[k] + hparam_c[k] - 1.0);
    for (k=0; k<num_class; k++)
      param_c[k] /= denom_c;
  }
  else if (smooth_c > 0) {
    denom_c = total_freq + smooth_c * num_class;
#pragma omp parallel for
    for (k=0; k<num_class; k++)
      param_c[k] = (expect_c[k] + smooth_c)/denom_c;
  }
  else {
#pragma omp parallel for
    for (k=0; k<num_class; k++)
      param_c[k] = expect_c[k]/total_freq;
  }

  if (mode_nominal) {
    if (mode_read_hparam) {
#pragma omp parallel for private(i,v,denom_w)
      for (k=0; k<num_class; k++) {
        for (v=0; v<tuple_size_d; v++) {
          denom_w = 0.0;
          for (i=0; i<num_word[v]; i++) {
            param_w[k][v][i] =
              expect_w[k][v][i] + num_miss_w[v] * param_w[k][v][i] +
              (hparam_w[k][v][i] - 1.0);
            denom_w += param_w[k][v][i];
          }
          for (i=0; i<num_word[v]; i++)
            param_w[k][v][i] /= denom_w;
        }
      }
    }
    else {
      for (v=0; v<tuple_size_d; v++) {
        if (smooth_w[v] > 0) {
#pragma omp parallel for private(i,denom_w)
          for (k=0; k<num_class; k++) {
            denom_w = 0.0;
            for (i=0; i<num_word[v]; i++) {
              param_w[k][v][i] =
                expect_w[k][v][i] +
                num_miss_w[v] * param_w[k][v][i] +
                smooth_w[v];
              denom_w += param_w[k][v][i];
            }
            for (i=0; i<num_word[v]; i++)
              param_w[k][v][i] /= denom_w;
          }
        }
        else {
#pragma omp parallel for private(i,denom_w)
          for (k=0; k<num_class; k++) {
            denom_w = 0.0;
            for (i=0; i<num_word[v]; i++) {
              param_w[k][v][i] =
                expect_w[k][v][i] + num_miss_w[v] * param_w[k][v][i];
              denom_w += param_w[k][v][i];
            }
            for (i=0; i<num_word[v]; i++)
              param_w[k][v][i] /= denom_w;
          }
        }
      }
    }
  }

  if (mode_gauss) {
    if (mode_smooth_gauss) {
      for (k=0; k<num_class; k++) {
        tau0 = get_tau(k);
        e_c_k = expect_c[k];

        for (u=0; u<gauss_size; u++) {
          n = gauss_attr_size[u];

          p_u = param_u[k][u];
          p_s = param_s[k][u];
          e_u = expect_u[k][u];
          e_s = expect_s[k][u];
          h_s = hparam_s[k][u];

          for (a=0; a<n; a++) {
            hparam_u0 = hparam_u[k][u][a];
            p_u[a] = (tau0 * hparam_u0 + e_u[a]) / (tau0 + e_c_k);
          }

          alpha_w0 = get_alpha_w(k,u,n);

          /* Compute the covariances in the upper triangle and then copy them
          to the lower triangle */
          for (a0=0; a0<n; a0++) {
            hparam_u0 = hparam_u[k][u][a0];
            for (a1=a0; a1<n; a1++) {
              hparam_u1 = hparam_u[k][u][a1];

              tmp_s =
                (h_s[a0][a1]
                 + e_s[a0][a1]
                 - p_u[a0] * e_u[a1]
                 - p_u[a1] * e_u[a0]
                 + e_c_k * p_u[a0] * p_u[a1]
                 + tau0 * (hparam_u0 - p_u[a0]) * (hparam_u1 - p_u[a1]))
                / ((alpha_w0 - n) + e_c_k);

              p_s[a0][a1] =
                (FABS(tmp_s) > SIGMA_LIMIT) ?
                tmp_s : ((a0 != a1 && tmp_s < 0) ? -SIGMA_LIMIT : SIGMA_LIMIT);
            }
          }
          copy_upper_to_lower(p_s);
        }
      }
    }
    else {
#pragma omp parallel for private(u,n,a,a0,a1,tmp_s)
      for (k=0; k<num_class; k++) {
        for (u=0; u<gauss_size; u++) {
          n = gauss_attr_size[u];

          for (a=0; a<n; a++)
            param_u[k][u][a] = expect_u[k][u][a] / expect_c[k];

          /* [Note]
           * - We only compute the covariances in the upper triangle
           * - Variances should be positive
           */
          for (a0=0; a0<n; a0++) {
            for (a1=a0; a1<n; a1++) {
              tmp_s =
                (expect_s[k][u][a0][a1] / expect_c[k])
                - param_u[k][u][a0] * param_u[k][u][a1];
              param_s[k][u][a0][a1] =
                (FABS(tmp_s) > SIGMA_LIMIT) ?
                tmp_s : ((a0 != a1 && tmp_s < 0) ? -SIGMA_LIMIT : SIGMA_LIMIT);
            }
          }
          copy_upper_to_lower(param_s[k][u]);
        }
      }
    }
  }

  if (mode_gauss) compute_param_precisions();

  if (mode_verbose) print_param(fp_log);
}

static void compute_case_member_prob_nolog(void)
{
  int k,e;

  compute_complete_prob();
  compute_incomplete_prob();
  compute_conditional_prob();

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      prob_mem_case[k][e] = prob_cond[k][e];

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      prob_compo_case[k][e] = prob_comp[k][e] / param_c[k];

  for (e=0; e<num_entry; e++)
    prob_case[e] = prob_incomp[e];
}

static void compute_case_member_prob_log(void)
{
  int k,e;

  compute_complete_prob();
  compute_incomplete_prob();
  compute_conditional_prob();

  /* Computing membership probabilities for cases */
#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      prob_mem_case[k][e] = prob_cond[k][e];

  /* Computing component probabilities for cases */
#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      prob_compo_case[k][e] = EXP(prob_comp[k][e] - param_c[k]);

#pragma omp parallel for
  for (e=0; e<num_entry; e++)
    prob_case[e] = EXP(prob_incomp[e]);
}

void compute_case_member_prob(void)
{
  if (mode_log_scale && mode_smooth_strict)
    compute_case_member_prob_log();
  else
    compute_case_member_prob_nolog();
}

static void compute_attr_member_prob_nolog(void)
{
  int k,v,i,e,u,n;
  VECTOR x,p_u;
  MATRIX p_s,p_p;
  int *obs,n_miss;
  FLOAT_T det;

  /* Computing membership probabilities for discrete attribute values */
  if (mode_nominal) {
#pragma omp parallel for private(i)
    for (v=0; v<tuple_size_d; v++)
      for (i=0; i<num_word[v]; i++)
        prob_w[v][i] = 0.0;

    /* Cannot immediately add a `pragma omp' statement here */
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          prob_w[v][i] += (prob_mem_w[k][v][i] = param_c[k] * param_w[k][v][i]);

#pragma omp parallel for private(v,i)
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          prob_compo_w[k][v][i] = param_w[k][v][i];

    /* normalization */
#pragma omp parallel for private(v,i)
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          prob_mem_w[k][v][i] /= prob_w[v][i];
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++)
      for (e=0; e<num_entry; e++)
        prob_g[u][e] = 0.0;

#pragma omp parallel for private(u,n,p_u,p_s,p_p,det,e,x,n_miss,obs)
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];

        p_u = param_u[k][u];
        p_s = param_s[k][u];
        p_p = param_p[k][u];
        det = param_sdet[k][u];

        for (e=0; e<num_entry; e++) {
          x = freq_list[e].gval[u];
          n_miss = freq_list[e].num_miss_g[u];

          prob_mem_g[k][u][e] = param_c[k];
          if (0 < n_miss && n_miss < n) {
            obs = freq_list[e].obs_g[u];
            prob_mem_g[k][u][e] *=
              (prob_compo_g[k][u][e] = gauss_marginal(x,p_u,p_s,obs));
          }
          else if (n_miss == 0)
            prob_mem_g[k][u][e] *=
              (prob_compo_g[k][u][e] = gauss_prec(x,p_u,p_s,p_p,det));
        }
      }
    }

#pragma omp parallel for private(e,k)
    for (u=0; u<gauss_size; u++) {
      for (e=0; e<num_entry; e++) {
        prob_g[u][e] = 0.0;
        for (k=0; k<num_class; k++)
          prob_g[u][e] += prob_mem_g[k][u][e];
      }
    }

#pragma omp parallel for private(u,e)
    for (k=0; k<num_class; k++)
      for (u=0; u<gauss_size; u++)
        for (e=0; e<num_entry; e++)
          prob_mem_g[k][u][e] /= prob_g[u][e];
  }

}

static void compute_attr_member_prob_log(void)
{
  int k,v,i,e,u,n;
  VECTOR x,p_u;
  MATRIX p_s,p_p;
  int *obs,n_miss;
  FLOAT_T det,p_max,sum;

  /* Computing membership probabilities for discrete attribute values */
  if (mode_nominal) {
#pragma omp parallel for private(v,i)
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          prob_mem_w[k][v][i] = LOG(param_c[k]) + LOG(param_w[k][v][i]);

#pragma omp parallel for private(i,k,p_max,sum)
    for (v=0; v<tuple_size_d; v++) {
      for (i=0; i<num_word[v]; i++) {
        p_max = 0.0;
        for (k=0; k<num_class; k++)
          if (k == 0 || prob_mem_w[k][v][i] > p_max)
            p_max = prob_mem_w[k][v][i];

        sum = 0.0;
        for (k=0; k<num_class; k++)
          sum += EXP(prob_mem_w[k][v][i] - p_max);

        prob_w[v][i] = p_max + LOG(sum);
      }
    }

    /* normalization */
#pragma omp parallel for private(v,i)
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          prob_mem_w[k][v][i] =
            EXP(prob_mem_w[k][v][i] - prob_w[v][i]); /* reused */


#pragma omp parallel for private(v,i)
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          prob_compo_w[k][v][i] = param_w[k][v][i];
  }

  if (mode_gauss) {
#pragma omp parallel for private(u,n,p_u,p_s,p_p,det,e,x,n_miss,obs)
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];

        p_u = param_u[k][u];
        p_s = param_s[k][u];
        p_p = param_p[k][u];
        det = param_sdet[k][u];

        for (e=0; e<num_entry; e++) {
          x = freq_list[e].gval[u];
          n_miss = freq_list[e].num_miss_g[u];

          prob_mem_g[k][u][e] = LOG(param_c[k]);
          if (0 < n_miss && n_miss < n) {
            obs = freq_list[e].obs_g[u];
            prob_mem_g[k][u][e] +=
              (prob_compo_g[k][u][e] = lgauss_marginal(x,p_u,p_s,obs));
          }
          else if (n_miss == 0)
            prob_mem_g[k][u][e] +=
              (prob_compo_g[k][u][e] = lgauss_prec(x,p_u,p_s,p_p,det));
        }
      }
    }

#pragma omp parallel for private(e,k,p_max,sum)
    for (u=0; u<gauss_size; u++) {
      for (e=0; e<num_entry; e++) {
        p_max = 0.0;
        for (k=0; k<num_class; k++)
          if (k == 0 || prob_mem_g[k][u][e] > p_max)
            p_max = prob_mem_g[k][u][e];

        sum = 0.0;
        for (k=0; k<num_class; k++)
          sum += EXP(prob_mem_g[k][u][e] - p_max);

        prob_g[u][e] = p_max + LOG(sum);
      }
    }

#pragma omp parallel for private(u,e)
    for (k=0; k<num_class; k++)
      for (u=0; u<gauss_size; u++)
        for (e=0; e<num_entry; e++)
          prob_mem_g[k][u][e] = EXP(prob_mem_g[k][u][e] - prob_g[u][e]);
  }

}

void compute_attr_member_prob(void)
{
  if (mode_log_scale && mode_smooth_strict)
    compute_attr_member_prob_log();
  else
    compute_attr_member_prob_nolog();
}

/*
 * body of the EM algorithm
 * - r: index of restarts
 * - m: starting index of EM iterations
 * - n: max number of iterations
 */
int em_iterate_body(int r, int m, int n, FLOAT_T itemp_init0)
{
  int mm;

  itemp = (mode_daem) ? itemp_init0 : 1.0;

  if (mode_daem && itemp < 1.0)
    compute_complete_prob_daem();
  else
    compute_complete_prob();
  compute_incomplete_prob();

  loglike[m] = compute_loglike()/itemp;
  /* lambda = log of a posteriori prob (MAP) or log-likelihood (ML) */
  if (mode_smooth) {
    log_prior[m] = compute_log_prior();
    lambda[m] = loglike[m] + log_prior[m];
  }
  else {
    lambda[m] = loglike[m];
  }
  if (mode_daem) log_itemp[m] = itemp;

  if (num_init > 1)
    print_msg("  [%d] #iters ",r);
  else
    print_msg("  #iters ");

  while (1) {  /* Loop for inverse temperatures */

    print_daem_message(m,itemp);
    mm = 0;

    /* Iterate E-step and M-step */
    while (1) {
      print_dot_message(m);

      compute_conditional_prob();
      compute_expect();
      update_param();
      if (mode_gauss && is_singular(r,m,n))
        break;  /* just discard the last update */

      m++;
      mm++;

      if (mode_daem && itemp < 1.0)
        compute_complete_prob_daem();
      else
        compute_complete_prob();
      compute_incomplete_prob();

      if (m >= ll_size) expand_loglike_all();
      loglike[m] = compute_loglike()/itemp;
      if (mode_smooth) {
        log_prior[m] = compute_log_prior();
        lambda[m] = loglike[m] + log_prior[m];
      }
      else {
        lambda[m] = loglike[m];
      }
      if (mode_daem) log_itemp[m] = itemp;

      if (is_decreased(r,m,n)) break;

      if (mode_daem && is_converged_daem(r,m,mm,n,itemp)) {
        break;
      }
      else if (!mode_daem && is_converged(r,m,n)) {
        break;
      }
    }

    if (!mode_daem && m >= n) break;

    if (itemp < 1.0) itemp *= itemp_rate;
    else break;
    if (itemp > 1.0) itemp = 1.0;
  }

  /* record the converged likelihood (a posteriori prob) */
  final_loglike = loglike[m];
  final_lambda = lambda[m];
  if (mode_smooth) final_log_prior = log_prior[m];

  /* return the index of the next iteration
   * (not the number of iterations conducted in this function)
   */
  return m;
}

/*
 * EM run(s) with some paticular number of classes
 * -- u is the index for searching the optimal number of classes
 */
void em_iterate(int u)
{
  int r,r_best=-1,n,n0,n_resume;
  int num_iterate,num_iterate_best=-1;
  unsigned int curr_seed=0;

  /* get the max number of EM iterations */
  n0 = (max_iterate > 0) ? max_iterate : MAX_MAX_ITERATE;
  if (num_init > 1) {
    n = (max_init_iterate > 0 && max_init_iterate < n0) ?
        max_init_iterate : n0;
  }
  else n = n0;
  n_resume = n0;

  for (r=0; r<num_init; r++) { /* random restarts */
    if (!mode_set_seed) {
      curr_seed = (unsigned int)random_int(); /* reset the random seeds */
      random_seed(curr_seed);
    }

    /* initialization step */
    initialize_param();

    /* run the body of EM, and get the number of iterations */
    num_iterate = em_iterate_body(r,0,n,itemp_init);

    /* record the results obtained by a random restart */
    log_seed[u][r] = (!mode_set_seed) ? curr_seed : 0;
    log_num_iterate[u][r] = num_iterate;
    log_final_lambda[u][r] = final_lambda;

    /* save the current estimate if we have the highest likelihood
     * (a positeriori prob)
     */
    if (r == 0 || final_lambda > final_lambda_best) {
      r_best = r;
      final_lambda_best = final_lambda;
      final_loglike_best = final_loglike;
      final_log_prior_best = final_log_prior;
      num_iterate_best = num_iterate;
      save_loglike_best(num_iterate_best);
      if (mode_daem) save_log_itemp_best(num_iterate_best);
      save_param_best();
      itemp_best = itemp;
    }
  }

  /* record the index of the current restart */
  log_r_best[u] = r_best;

  /* restore the parameters with the highest likelihood
   * (a posteriori prob), etc.
   */
  final_lambda = final_lambda_best;
  final_loglike = final_loglike_best;
  final_log_prior = final_log_prior_best;

  restore_param_best();
  restore_loglike_best(num_iterate_best);

  if (mode_daem) restore_log_itemp_best(num_iterate_best);
  num_iterate = num_iterate_best;

  /* continue the EM iterations */
  if (num_init > 1) {
    print_msg("  <<Resumed best parameter set #%d>>\n",r_best);
    num_iterate = em_iterate_body(r_best,num_iterate,n_resume,itemp_best);
  }

  /* output the likelihood (or the a posteriori prob) */
  print_loglike(fp_log,num_class,num_iterate);
}

void em_main(void)
{
  int u,num_search;
  FLOAT_T ic_opt=0.0;
  FILE *fp_param,*fp_memp;

  /* Read the occurrence data first */
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_read = current_sec();
  stime_lap_read = elapsed_sec();
  read_freq();
  etime_read = current_sec();
  etime_lap_read = elapsed_sec();
#else
  read_freq();
#endif

  /**** Run the EM algorithm: ****/

  num_search = (mode_search_num_class) ? ((kmax-kmin)/kstep + 1) : 1;
  alloc_ic(num_search);
  alloc_seed();

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_em = current_sec();
  stime_lap_em = elapsed_sec();
#endif

  /* we attempt to search for the optimal number of classes: */
  if (mode_search_num_class && !mode_read_hparam) {
    print_size(fp_log);

    u = 0;

    for (num_class = kmin; num_class <= kmax; num_class += kstep) {
      print_msg("|Classes| = %d:\n",num_class);

      /* (re-)set the smoothing constants according to the number of classes*/
      set_smooth_auto();

      /* allocate memory areas for EM */
      alloc_param();
      alloc_param_best();
      alloc_hparam_gauss();
      initialize_hparam_gauss();
      alloc_prob();
      alloc_expect();
      ll_size = MAX_MAX_ITERATE;
      alloc_loglike();
      if (mode_daem) alloc_log_itemp();

      /* run the body part of the EM algorithm */
      if (num_sm > 0 && num_class > 2)
        smem_iterate(u);
      else
        em_iterate(u);

      /* compute the model scores */
      bic[u] = compute_bic();
      if (mode_smooth_strict) autoclass[u] = compute_cs();
      switch (ic_type) {
      case BIC:
        ic[u] = bic[u];
        break;
      case AUTOCLASS:
        ic[u] = autoclass[u];
        break;
      default:
        die("Internal error: Unsupported information criterion\n");
      }

      release_param_best();
      release_prob();
      release_expect();
      release_loglike();
      if (mode_daem) release_log_itemp();

      /* now we have the (temporarily) optimal number of classes */
      if (num_class == kmin || ic[u] > ic_opt) {
        print_msg("  %s = %.Nf (temporarily optimal)\n",ic_type_str,ic[u]);
        if (num_class > kmin) release_param_opt();
        num_class_opt = num_class;
        ic_opt = ic[u];

        save_param_opt();  /* save the current parameters */
      }
      else
        print_msg("  %s = %.Nf\n",ic_type_str,ic[u]);

      release_param();
      release_hparam_gauss();

      u++;
    }
    print_msg("Optimal |Classes| = %d\n",num_class_opt);

    /* restore the parameters with the optimal number of classes */
    num_class = num_class_opt;
    restore_param_opt();
    release_param_opt();

    print_ic(fp_log);
    print_seed(fp_log);
  }
  /* we do not try to find the optimal number of classes */
  else {
    if (mode_read_hparam) {
      read_hparam();   /* alloc_hparam{,_aux,_gauss}() will be called  */
      check_read_hparam();        /* check if mode_smooth_strict holds */
    }
    else {
      set_smooth_auto();
      alloc_hparam_gauss();
      initialize_hparam_gauss();
    }
    print_size(fp_log);

    print_msg("#classes = %d:\n",num_class);

    alloc_param();
    ll_size = MAX_MAX_ITERATE;
    alloc_loglike();
    if (mode_daem) alloc_log_itemp();
    alloc_prob();
    alloc_expect();
    alloc_param_best();

    if (num_sm > 0)
      smem_iterate(0);
    else
      em_iterate(0);

    /* compute the model scores */
    bic[0] = compute_bic();
    if (mode_smooth_strict) autoclass[0] = compute_cs();
    switch (ic_type) {
    case BIC:
      ic[0] = bic[0];
      break;
    case AUTOCLASS:
      ic[0] = autoclass[0];
      break;
    default:
      die("Internal error: Unsupported information criterion\n");
    }

    if (mode_read_hparam) release_hparam();
    release_param_best();
    release_prob();
    release_expect();
    release_loglike();
    if (mode_daem) release_log_itemp();

    print_msg("  %s = %.Nf\n",ic_type_str,ic[0]);

    print_ic(fp_log);
    print_seed(fp_log);
  }

  release_ic();
  release_seed();

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  etime_em = current_sec();
  etime_lap_em = elapsed_sec();
#endif

  if (mode_verbose) print_param(fp_log);

  /* at this point, num_class, param_c[] and param_w[][] store the values
   * with the optimal number of classes
   */

  /* record the estimated parameters */
  fp_param = fopen(fname_param,"wb");
  if (fp_param == NULL) die("Couldn't open %s\n", fname_param);
  print_param(fp_param);
  fclose(fp_param);

  /* compute and record the membership probabilities */
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  if (mode_memp) stime_memp = current_sec();
  if (mode_memp) stime_lap_memp = elapsed_sec();
#endif

  if (mode_memp || mode_cluster || mode_rank || mode_dissim) {
    alloc_prob_mem();
    alloc_prob();
    compute_case_member_prob();
    compute_attr_member_prob();
    release_prob();
  }

  if (mode_memp) {
    fp_memp = fopen(fname_memp,"wb");
    if (fp_memp == NULL) die("Couldn't open %s\n", fname_memp);
    print_case_member_prob(fp_memp);
    if (mode_show_attr_memp) print_attr_member_prob(fp_memp);
    fclose(fp_memp);
  }

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  if (mode_memp) etime_memp = current_sec();
  if (mode_memp) etime_lap_memp = elapsed_sec();
#endif

  /* clustering & evaluation */
  if (mode_cluster) cluster_main();

  /* relevance analysis */
  if (mode_rank) rank_main();

  /* computing dissimilarities */
  if (mode_dissim) dissim_main();

  release_param();
  release_hparam_gauss();
  if (mode_memp || mode_cluster || mode_rank || mode_dissim)
    release_prob_mem();

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  print_time(fp_log); /* record the computation time */
#endif
}

/*
 * we do not run the EM algorithm, but read the parameters obtained
 * by a previous run of nbc
 */
void no_em_main(void)
{
  FILE *fp_memp;

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_read = current_sec();
  stime_lap_read = elapsed_sec();
#endif

  read_param(); /** read the parameters (alloc_param() is called) **/
  if (mode_cluster || mode_rank) read_freq();

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  etime_read = current_sec();
  etime_lap_read = elapsed_sec();
#endif

  set_smooth_auto();

  print_size(fp_log);

  /* computing membership probabilities */
  if (mode_memp) {
    print_msg("Computing membership distributions...");
    fflush(fp_msg);
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
    stime_memp = current_sec();
    stime_lap_memp = elapsed_sec();
#endif
  }

  if (mode_memp || mode_cluster || mode_rank || mode_dissim) {
    alloc_prob_mem();
    alloc_prob();
    compute_case_member_prob();
    compute_attr_member_prob();
    release_prob();
  }

  if (mode_memp) {
    fp_memp = fopen(fname_memp,"wb");
    if (fp_memp == NULL) die("Couldn't open %s\n", fname_memp);
    print_case_member_prob(fp_memp);
    if (mode_show_attr_memp) print_attr_member_prob(fp_memp);
    fclose(fp_memp);
  }

  if (mode_memp) {
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
    etime_memp = current_sec();
    etime_lap_memp = elapsed_sec();
#endif
    print_msg("done\n");
  }

  /* clustering and evaluation */
  if (mode_cluster) cluster_main();

  /* relevance analysis */
  if (mode_rank) rank_main();

  /* computing dissimilarities */
  if (mode_dissim) dissim_main();

  /* cleaning-up */
  release_param();
  if (mode_memp || mode_cluster || mode_rank || mode_dissim)
    release_prob_mem();

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  print_time(fp_log);
#endif
}
