/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

static int hash(HASHTABLE *tp, char *s)
{
  unsigned int v;

  for (v = 0; *s != '\0'; s++)
    v = v * HASH_RADIX + *s;

  return (int)(v % tp->size);
}

#if 0
static void print_hashtable(HASHTABLE *tp)
{
  HASH_NODE_PTR q;
  int i;

  print_msg("---- hashtable ----\n");
  print_msg("size = %d\n", tp->size);
  print_msg("used = %d\n", tp->used);

  for (i=0; i<tp->size; i++) {
    print_msg("[%d]",i);
    for (q = tp->table[i]; q != NULL; q = q->next) {
      print_msg("[%s (%d)]",q->key,q->count);
    }
    print_msg("\n");
  }
}
#endif

static void expand_hashtable(HASHTABLE *tp)
{
  HASH_NODE_PTR *old_table = tp->table;
  int old_size             = tp->size;
  int i;
  HASH_NODE_PTR *p, q, r;

  tp->size   = 2 * tp->size + 1;
  tp->table  = (HASH_NODE_PTR *)xmalloc(sizeof(HASH_NODE_PTR)*tp->size);
  tp->used   = 0;

  for (i=0; i<tp->size; i++)
    tp->table[i] = NULL;

  /* reconnect: we access the old entries similarly to release_hashtable()  */
  for (i=0; i<old_size; i++) {
    for (q = old_table[i]; q != NULL; q = r) {
      r = q->next;

      p = tp->table + hash(tp, q->key);  /* rehash */

      if ((*p) == NULL) {
        tp->used++;
      }
      else {
        /* each key is unique, so we immediately go to the end of the chain */
        while ((*p) != NULL)
          p = &((*p)->next);
      }

      *p = q;
      q->next = NULL;
    }
  }

  xfree(old_table);
}

void init_hashtable(HASHTABLE *tp)
{
  int i;

  tp->serial_id = 0;
  tp->size      = HASH_SIZE;
  tp->table     = (HASH_NODE_PTR *)xmalloc(sizeof(HASH_NODE_PTR)*tp->size);
  tp->used      = 0;

  for (i=0; i<tp->size; i++) tp->table[i] = NULL;
}

void release_hashtable(HASHTABLE *tp)
{
  int i;
  HASH_NODE_PTR q,r;

  for (i=0; i<tp->size; i++) {
    for (q = tp->table[i]; q != NULL; q = r) {
      r = q->next;
      xfree(q->key);
      xfree(q);
    }
  }

  xfree(tp->table);
}

int get_cardinality(HASHTABLE *tp)
{
  return tp->serial_id;
}

int lookup_hashtable(HASHTABLE *tp, char *entry)
{
  HASH_NODE_PTR q;

  q = tp->table[hash(tp, entry)];

  while (q != NULL) {
    if (strcmp(entry, q->key) == 0)
      return q->id;
    else
      q = q->next;
  }

  return -1;
}

int register_hashtable(HASHTABLE *tp, char *entry)
{
  HASH_NODE_PTR *p, q;
  double usage;

  p = tp->table + hash(tp, entry);

  if ((*p) == NULL) {
    tp->used++;
  }
  else {
    while ((*p) != NULL) {
      if (strcmp(entry, (*p)->key) == 0) {
        (*p)->count++;
        return (*p)->id;
      }
      else
        p = &((*p)->next);
    }
  }

  q = (struct HASH_NODE *)xmalloc(sizeof(struct HASH_NODE));
  q->key = alloc_str(strlen(entry) + 1);

  strncpy(q->key,entry,(strlen(entry) + 1));
  q->id = tp->serial_id;
  q->count = 1;
  q->next = NULL;

  tp->serial_id++;
  *p = q;

  usage = tp->used/(double)tp->size;
  if (usage > HASH_EXPAND)
    expand_hashtable(tp);

  return q->id;
}

char **get_keys(HASHTABLE *tp)
{
  int i;
  HASH_NODE_PTR q;
  char **el;

  el = (char **)xmalloc(sizeof(char *)*tp->serial_id);

  for (i=0; i<tp->size; i++) {
    q = tp->table[i];
    while (q != NULL) {
      el[q->id] = q->key;  /* not copied */
      q = q->next;
    }
  }

  return el;  /* should be freed later */
}

/* get the most frequent key in the hashtable */
void get_mode_key(HASHTABLE *tp, char **key, int *count)
{
  char *key_max;
  int i,count_max;
  HASH_NODE_PTR q;

  count_max = -1;
  key_max = NULL;
  for (i=0; i<tp->size; i++) {
    q = tp->table[i];
    while (q != NULL) {
      if (q->count > count_max ||
          (q->count == count_max && strcmp(q->key, key_max) < 0)) {
        key_max = q->key;  /* not copied */
        count_max = q->count;
      }
      q = q->next;
    }
  }

  *key = key_max;
  *count = count_max;
}

static int compare_hash_entry_by_key(const void *a, const void *b)
{
  HASH_ENTRY_PTR entry_a = (HASH_ENTRY_PTR)a;
  HASH_ENTRY_PTR entry_b = (HASH_ENTRY_PTR)b;

  return strcmp(entry_a->key, entry_b->key);  /* in ascending order */
}

static int compare_hash_entry_by_count(const void *a, const void *b)
{
  HASH_ENTRY_PTR entry_a = (HASH_ENTRY_PTR)a;
  HASH_ENTRY_PTR entry_b = (HASH_ENTRY_PTR)b;

  return (entry_b->count - entry_a->count);  /* in descending order */
}

/* get arrays of entries in the hashtable */
HASH_ENTRY_PTR get_entries(HASHTABLE *tp)
{
  HASH_ENTRY_PTR entries;
  HASH_NODE_PTR q;
  int c,i,id;

  c = get_cardinality(tp);
  entries = (HASH_ENTRY_PTR)xmalloc(sizeof(struct HASH_ENTRY)*c);

  for (i=0; i<tp->size; i++) {
    for (q = tp->table[i]; q != NULL; q = q->next) {
      id = q->id;
      entries[id].key   = q->key;  /* not copied */
      entries[id].id    = q->id;
      entries[id].count = q->count;
    }
  }

  return entries;   /* should be freed later */
}

/* get an array of entries in the table, sorted in the lexicographical order */
HASH_ENTRY_PTR get_entries_sorted_by_key(HASHTABLE *tp)
{
  HASH_ENTRY_PTR entries;
  int c;

  entries = get_entries(tp);
  c = get_cardinality(tp);

  qsort(entries,c,sizeof(struct HASH_ENTRY),
        (int(*)(const void *,const void *))compare_hash_entry_by_key);

  return entries;   /* should be freed later */
}

/* get an array of entries in the table, sorted by the occurrences */
HASH_ENTRY_PTR get_entries_sorted_by_count(HASHTABLE *tp)
{
  HASH_ENTRY_PTR entries;
  int c;

  entries = get_entries(tp);
  c = get_cardinality(tp);

  qsort(entries,c,sizeof(struct HASH_ENTRY),
        (int(*)(const void *,const void *))compare_hash_entry_by_count);

  return entries;  /* should be freed later */
}
