/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"
#ifdef VNBC
#  include "vnbc_def.h"
#  include "vnbc_ext.h"
#endif

/* Get an estimation of means and (co-)variances for numeric attributes
 * with no consideration about the latent classes
 *
 * -- For the data including missing cells, it should be better to run
 *    the EM algorithm, but here we only make a rough estimation
 */
static void compute_sample_stats(void)
{
  int u,a,a0,a1,n,e,f;
  FLOAT_T total_miss,det;
  FREQ_PTR f_ent;
  int miss0, miss1;

  sample_u_raw = (VECTOR *)xmalloc(sizeof(VECTOR)*gauss_size);
  sample_s_raw = (MATRIX *)xmalloc(sizeof(MATRIX)*gauss_size);

  for (u=0; u<gauss_size; u++) {
    n = gauss_attr_size[u];
    sample_u_raw[u] = alloc_clear_vector(n);
    sample_s_raw[u] = alloc_clear_matrix(n,n);
  }

  for (u=0; u<gauss_size; u++) {
    n = gauss_attr_size[u];

    for (a=0; a<n; a++) {

      total_miss = 0;
      for (e=0; e<num_entry; e++) {
        f_ent = &freq_list[e];
        f = f_ent->freq;
        if (f_ent->miss_g[u][a])
          total_miss += f;
        else
          sample_u_raw[u][a] += f * f_ent->gval[u][a];
      }

      sample_u_raw[u][a] /= (total_freq - total_miss);
    }
  }

  for (u=0; u<gauss_size; u++) {
    n = gauss_attr_size[u];

    for (a0=0; a0<n; a0++) {
      for (a1=a0; a1<n; a1++) {

        total_miss = 0;
        for (e=0; e<num_entry; e++) {
          f_ent = &freq_list[e];
          f = f_ent->freq;
          miss0 = f_ent->miss_g[u][a0];
          miss1 = f_ent->miss_g[u][a1];
          if (miss0 && miss1) {
            total_miss += f;
          }
          else if (miss0 || miss1) {
            /* We add nothing to sample_s_raw[u][a0][a1] nor total_miss,
             * by replacing with missing values by means, i.e. assuming that
             * f_ent->gval[u][a0] = sample_u_raw[u][a0] or
             * f_ent->gval[u][a1] = sample_u_raw[u][a1] (see the next case).
             */
          }
          else {
            sample_s_raw[u][a0][a1] +=
              f * (f_ent->gval[u][a0] - sample_u_raw[u][a0]) *
              (f_ent->gval[u][a1] - sample_u_raw[u][a1]);
          }
        }

        sample_s_raw[u][a0][a1] /= (total_freq - total_miss) - 1;
      }
    }
    copy_upper_to_lower(sample_s_raw[u]);
  }

  /* Check the singularity of the sample covariances by using
   * LU-decomposition based routine
   */
#ifdef VNBC
  if (mode_gauss && mode_smooth_s != SMOOTH_VAR_CONST) {
    for (u=0; u<gauss_size; u++) {
      det = determinant(sample_s_raw[u]); /* LU decomposition based */
      if (FABS(det) < VERY_SMALL_FLOAT && mode_smooth_s == SMOOTH_VAR_SAMPLE)
        die("Sample covariance matrix is singular; "
            "please specify a diagonal covariance matrix by '%s const:Var'",
            LOPTSTR_VB(O_VB_SMOOTH_VAR));
    }
  }
#else
  if (mode_gauss && mode_smooth_gauss && mode_smooth_s != SMOOTH_VAR_CONST) {
    for (u=0; u<gauss_size; u++) {
      det = determinant(sample_s_raw[u]); /* LU decomposition based */
      if (FABS(det) < VERY_SMALL_FLOAT)
        die("Sample covariance matrix is singular; "
            "please specify a diagonal covariance matrix by '%s const:Var'",
            LOPTSTR(O_SMOOTH_VAR));
    }
  }
#endif
}

static void compute_regular_stat(void)
{
  int u,a,a0,a1,n,e,f,i;
  int total_miss,m,miss,miss0,miss1;
  FLOAT_T det,x,x0,x1;
  FREQ_PTR f_ent;
  FLOAT_T *data;
  int *freq;

  regular_u   = (VECTOR *)xmalloc(sizeof(VECTOR)*gauss_size);
  regular_s   = (MATRIX *)xmalloc(sizeof(MATRIX)*gauss_size);
  regular_min = (VECTOR *)xmalloc(sizeof(VECTOR)*gauss_size);
  regular_max = (VECTOR *)xmalloc(sizeof(VECTOR)*gauss_size);

  for (u=0; u<gauss_size; u++) {
    n = gauss_attr_size[u];
    regular_u[u]   = alloc_clear_vector(n);
    regular_s[u]   = alloc_clear_matrix(n,n);
    regular_min[u] = alloc_clear_vector(n);
    regular_max[u] = alloc_clear_vector(n);
  }

  for (u=0; u<gauss_size; u++) {
    n = gauss_attr_size[u];

    for (a=0; a<n; a++) {

      m = 0;
      for (e=0; e<num_entry; e++) {
        f_ent = &freq_list[e];
        f = f_ent->freq;
        if (f_ent->miss_g[u][a]) continue;

        m++;
      }

      data = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*m);
      freq = (int *)xmalloc(sizeof(int)*m);

      i = 0;
      for (e=0; e<num_entry; e++) {
        f_ent = &freq_list[e];
        if (f_ent->miss_g[u][a]) continue;

        data[i] = f_ent->gval[u][a];
        freq[i] = f_ent->freq;
        i++;
      }

      regular_min[u][a] = get_quantile_count(q_regular_min,m,data,freq);
      regular_max[u][a] = get_quantile_count(q_regular_max,m,data,freq);

      xfree(data);
      xfree(freq);
    }
  }

  for (u=0; u<gauss_size; u++) {
    n = gauss_attr_size[u];

    for (a=0; a<n; a++) {

      total_miss = 0;
      for (e=0; e<num_entry; e++) {
        f_ent = &freq_list[e];
        f = f_ent->freq;
        x = f_ent->gval[u][a];
        miss = f_ent->miss_g[u][a];
        if (miss ||	x < regular_min[u][a] || x > regular_max[u][a])
          total_miss += f;
        else
          regular_u[u][a] += f * x;
      }

      regular_u[u][a] /= (total_freq - total_miss);
    }
  }

  for (u=0; u<gauss_size; u++) {
    n = gauss_attr_size[u];

    for (a0=0; a0<n; a0++) {
      for (a1=a0; a1<n; a1++) {

        total_miss = 0;
        for (e=0; e<num_entry; e++) {
          f_ent = &freq_list[e];
          f = f_ent->freq;
          miss0 = f_ent->miss_g[u][a0];
          miss1 = f_ent->miss_g[u][a1];
          x0 = f_ent->gval[u][a0];
          x1 = f_ent->gval[u][a1];

          if (miss0 && miss1) {
            total_miss += f;
          }
          else if (miss0 || miss1) {
            if (x0 < regular_min[u][a0] || x0 > regular_max[u][a0] ||
                x1 < regular_min[u][a1] || x0 > regular_max[u][a1])
              total_miss += f;
          }
          else {
            regular_s[u][a0][a1] +=
              f * (x0 - regular_u[u][a0]) * (x1 - regular_u[u][a1]);
            m++;
          }
        }

        regular_s[u][a0][a1] /= (total_freq - total_miss) - 1;
      }
    }
    copy_upper_to_lower(regular_s[u]);
  }

  /* Check the singularity of the sample covariances by using
   * LU-decomposition based routine
   */
#ifdef VNBC
  if (mode_gauss && mode_smooth_gauss && mode_smooth_s != SMOOTH_VAR_CONST) {
    for (u=0; u<gauss_size; u++) {
      det = determinant(regular_s[u]); /* LU decomposition based */
      if (FABS(det) < VERY_SMALL_FLOAT && mode_smooth_s == SMOOTH_VAR_SAMPLE)
        die("Sample covariance matrix is singular; "
            "please specify a diagonal covariance matrix by '%s const:Var'",
            LOPTSTR_VB(O_VB_SMOOTH_VAR));
    }
  }
#else
  if (mode_gauss && mode_smooth_s != SMOOTH_VAR_CONST) {
    for (u=0; u<gauss_size; u++) {
      det = determinant(regular_s[u]); /* LU decomposition based */
      if (FABS(det) < VERY_SMALL_FLOAT && mode_smooth_s == SMOOTH_VAR_SAMPLE)
        die("Sample covariance matrix is singular; "
            "please specify a diagonal covariance matrix by '%s const:Var'",
            LOPTSTR(O_SMOOTH_VAR));
    }
  }
#endif
}

static void compute_basic_stats(void)
{
  int e,v,v_d,u,a,i;
  int count,f,tried,m;
  int *freq;
  char *key;
  FREQ_PTR f_ent;
  FLOAT_T *data,max,min;

  attr_stat = (struct STAT *)xmalloc(sizeof(struct STAT)*tuple_size);

  if (mode_nominal) {
    for (v=0; v<tuple_size; v++) {
      if (!is_attr_disc(v)) continue;

      v_d = attr_info[v].local_id;

      attr_stat[v].num_val = num_word[v_d];

      get_mode_key(&ht[v_d],&key,&count);
      attr_stat[v].mode_val = key;
      attr_stat[v].mode_count = count;
    }
  }

  if (mode_gauss) {
    for (v=0; v<tuple_size; v++) {
      if (!is_attr_cont(v)) continue;

      u = attr_info[v].gauss_id;
      a = attr_info[v].local_id;

      attr_stat[v].mean = sample_u_raw[u][a];
      attr_stat[v].var  = sample_s_raw[u][a][a];
      attr_stat[v].std  = SQRT(sample_s_raw[u][a][a]);

      max = -1;
      min = -1;
      tried = 0;
      m = 0;
      for (e=0; e<num_entry; e++) {
        f_ent = &freq_list[e];
        f = f_ent->freq;
        if (f_ent->miss_g[u][a]) continue;

        if (!tried || f_ent->gval[u][a] > max) max = f_ent->gval[u][a];
        if (!tried || f_ent->gval[u][a] < min) min = f_ent->gval[u][a];

        m++;
        tried = 1;
      }

      attr_stat[v].max = max;
      attr_stat[v].min = min;

      data = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*m);
      freq = (int *)xmalloc(sizeof(int)*m);

      i = 0;
      for (e=0; e<num_entry; e++) {
        f_ent = &freq_list[e];
        if (f_ent->miss_g[u][a]) continue;

        data[i] = f_ent->gval[u][a];
        freq[i] = f_ent->freq;
        i++;
      }

      attr_stat[v].p0   = attr_stat[v].min;
      attr_stat[v].p5   = get_quantile_count(0.05,m,data,freq);
      attr_stat[v].p10  = get_quantile_count(0.10,m,data,freq);
      attr_stat[v].p25  = get_quantile_count(0.25,m,data,freq);
      attr_stat[v].p50  = get_quantile_count(0.50,m,data,freq);
      attr_stat[v].p75  = get_quantile_count(0.75,m,data,freq);
      attr_stat[v].p90  = get_quantile_count(0.90,m,data,freq);
      attr_stat[v].p95  = get_quantile_count(0.95,m,data,freq);
      attr_stat[v].p100 = attr_stat[v].max;

      attr_stat[v].median = attr_stat[v].p50;

      xfree(data);
      xfree(freq);
    }
  }
}

void compute_stats(void)
{
  if (mode_gauss) {
    compute_sample_stats();
    compute_regular_stat();
  }
  compute_basic_stats();

  if (mode_regular_stat) {
    sample_u = regular_u;
    sample_s = regular_s;
  }
  else {
    sample_u = sample_u_raw;
    sample_s = sample_s_raw;
  }

  print_basic_stats(fp_log);
  if (mode_gauss) {
    print_sample_stats(fp_log);
    print_regular_stat(fp_log);
  }
}
