/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"
#include "vnbc_def.h"
#include "vnbc_ext.h"

/* [Note] The returned probability is in log-scale */
FLOAT_T compute_predict_gauss(int k, int u, int e)
{
  int n,n_miss;
  int *obs;
  FLOAT_T p,e_c,f,c,tau0;
  MATRIX m;
  MATRIX **h_s;
  VECTOR **h_u;
  VECTOR x;

  n = gauss_attr_size[u];
  n_miss = freq_list[e].num_miss_g[u];

  x = freq_list[e].gval[u];

  if (mode_em) {
    e_c = expect_c[k];
    h_u = hparam_u_upd;
    h_s = hparam_s_upd;
    tau0 = tau + e_c;
    f = get_alpha_w(n) + e_c + 1 - n;
  }
  else {
    e_c = 0.0;
    h_u = hparam_u;
    h_s = hparam_s;
    tau0 = hparam_tau[k];
    f = hparam_alpha_w[k][u] + 1 - n;
  }

  m = dup_matrix(h_s[k][u]);
  c = (tau0 + 1) / (tau0 * f);
  mul_const_matrix(m,c);

  if (n_miss == 0) {
    p = lstudents_t(x,h_u[k][u],m,f); /* log-scale */
  }
  else if (0 < n_miss && n_miss < n) {
    obs = freq_list[e].obs_g[u];
    p = lstudents_t_marginal(x,h_u[k][u],m,f,obs);
  }
  else
    p = 0;

  release_matrix(m);

  return p;
}

/* We need log-scale computation */
void compute_case_member_prob(void)
{
  int k,v,e,u;
  FLOAT_T p_c,p_ic,p_c_max,sum;

#pragma omp parallel for private(v,p_ic,k,p_c,u,p_c_max,sum)
  for (e=0; e<num_entry; e++) {

    if (total_num_miss_w > 0 && freq_list[e].num_miss_w > 0) {
      p_c_max = 0.0;
      for (k=0; k<num_class; k++) {
        p_c = LOG(param_c[k]);

        if (mode_nominal) {
          for (v=0; v<tuple_size_d; v++)
            if (!freq_list[e].miss_w[v])
              p_c += LOG(param_w[k][v][freq_list[e].id[v]]);
        }

        if (mode_gauss) {
          for (u=0; u<gauss_size; u++)
            p_c += compute_predict_gauss(k,u,e);
        }

        prob_mem_case[k][e] = p_c;
        if (k == 0 || p_c > p_c_max) p_c_max = p_c;
      }

      sum = 0.0;
      for (k=0; k<num_class; k++)
        sum += EXP(prob_mem_case[k][e] - p_c_max);

      p_ic = p_c_max + LOG(sum);
    }
    else {
      p_c_max = 0.0;
      for (k=0; k<num_class; k++) {
        p_c = LOG(param_c[k]);

        if (mode_nominal) {
          for (v=0; v<tuple_size_d; v++)
            p_c += LOG(param_w[k][v][freq_list[e].id[v]]);
        }

        if (mode_gauss) {
          for (u=0; u<gauss_size; u++)
            p_c += compute_predict_gauss(k,u,e);
        }

        prob_mem_case[k][e] = p_c;
        if (k == 0 || p_c > p_c_max) p_c_max = p_c;
      }

      sum = 0.0;
      for (k=0; k<num_class; k++)
        sum += EXP(prob_mem_case[k][e] - p_c_max);

      p_ic = p_c_max + LOG(sum);
    }

    for (k=0; k<num_class; k++) {
      prob_compo_case[k][e] = EXP(prob_mem_case[k][e] - LOG(param_c[k]));
      prob_mem_case[k][e] = EXP(prob_mem_case[k][e] - p_ic);  /* Update */
    }
    prob_case[e] = EXP(p_ic);
  }

  if (mode_verbose) print_case_member_prob(fp_log);
}

void compute_attr_member_prob(void)
{
  int k,v,i,u,e;
  FLOAT_T p_max,sum;

  if (mode_nominal) {
#pragma omp parallel for private(i)
    for (v=0; v<tuple_size_d; v++)
      for (i=0; i<num_word[v]; i++)
        prob_w[v][i] = 0.0;

    /* Cannot immediately add a `pragma omp' statement here */
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          prob_w[v][i] += (prob_mem_w[k][v][i] = param_c[k] * param_w[k][v][i]);

    /* Normalization */
#pragma omp parallel for private(v,i)
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          prob_mem_w[k][v][i] /= prob_w[v][i];

#pragma omp parallel for private(v,i)
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          prob_compo_w[k][v][i] = param_w[k][v][i];
  }

  if (mode_gauss) {
#pragma omp parallel for private(e,k,p_max,sum)
    for (u=0; u<gauss_size; u++) {
      for (e=0; e<num_entry; e++) {
        p_max = 0.0;
        for (k=0; k<num_class; k++) {
          prob_mem_g[k][u][e] =
            LOG(param_c[k]) + compute_predict_gauss(k,u,e);

          if (k == 0 || prob_mem_g[k][u][e] > p_max)
            p_max = prob_mem_g[k][u][e];
        }

        sum = 0.0;
        for (k=0; k<num_class; k++)
          sum += EXP(prob_mem_g[k][u][e] - p_max);

        prob_g[u][e] = p_max + LOG(sum);
      }
    }

#pragma omp parallel for private(u,e)
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        for (e=0; e<num_entry; e++) {
          prob_compo_g[k][u][e] = EXP(prob_mem_g[k][u][e] - LOG(param_c[k]));
          prob_mem_g[k][u][e] = EXP(prob_mem_g[k][u][e] - prob_g[u][e]);
        }
      }
    }
    for (u=0; u<gauss_size; u++)
      for (e=0; e<num_entry; e++)
        prob_g[u][e] = EXP(prob_g[u][e]);
  }

  if (mode_verbose) print_attr_member_prob(fp_log);
}

void vbem_main(void)
{
  int u,num_search;
  FLOAT_T ic_opt=0;
  FILE *fp_hparam,*fp_memp,*fp_predict;

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_read = current_sec();
  stime_lap_read = elapsed_sec();
  read_freq();
  etime_read = current_sec();
  etime_lap_read = elapsed_sec();
#else
  read_freq();
#endif

  num_search = (mode_search_num_class) ? (kmax-kmin/kstep + 1) : 1;
  alloc_ic(num_search);
  alloc_seed();
  alloc_seed_vb();

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_em = current_sec();
  stime_lap_em = elapsed_sec();
#endif

  print_size(fp_log);

  if (mode_search_num_class) {
    u = 0;

    for (num_class = kmin; num_class <= kmax; num_class += kstep) {
      print_msg("|Classes| = %d:\n",num_class);

      set_smooth_auto();

      alloc_hparam();
      alloc_hparam_gauss();
      initialize_hparam_gauss();
      alloc_hparam_upd();
      alloc_hparam_best();

      alloc_pi();
      alloc_prob();
      alloc_expect();
      free_energy_size = MAX_MAX_ITERATE;
      alloc_free_energy();
      alloc_free_energy_best();
      if (mode_daem) alloc_log_itemp_vb();

      if (num_sm > 0 && num_class > 2) {
        alloc_prob_mem();
        vbsmem_iterate(u);
        release_prob_mem();
      }
      else vbem_iterate(u);

      compute_pi();
      compute_complete_prob();
      compute_incomplete_prob();
      ic[u] = compute_free_energy();

      release_hparam_best();
      release_free_energy_best();
      if (mode_daem) release_log_itemp();

      if (num_class == kmin || ic[u] > ic_opt) {
        print_msg("  Variational Free Energy = %.Nf (temporarily optimal)\n",
                  ic[u]);
        if (num_class > kmin) release_hparam_opt();
        num_class_opt = num_class;
        ic_opt = ic[u];

        save_hparam_opt();
      }
      else
        print_msg("  Variational Free Energy = %.Nf\n",ic[u]);

      release_hparam();
      release_hparam_gauss();
      release_hparam_upd();

      u++;
    }
    print_msg("Optimal |Classes| = %d\n",num_class_opt);

    num_class = num_class_opt;
    restore_hparam_opt();
    release_hparam_opt();

    print_ic(fp_log);
    print_seed(fp_log);
  }
  else {
    print_msg("|Classes| = %d:\n",num_class);

    set_smooth_auto();

    alloc_hparam();
    alloc_hparam_gauss();
    initialize_hparam_gauss();
    alloc_hparam_upd();

    free_energy_size = MAX_MAX_ITERATE;
    alloc_free_energy();
    alloc_free_energy_best();
    if (mode_daem) alloc_log_itemp_vb();
    alloc_pi();
    alloc_prob();
    alloc_expect();
    alloc_hparam_best();

    if (num_sm > 0 && num_class > 2) {
      alloc_prob_mem();
      vbsmem_iterate(0);
      release_prob_mem();
    }
    else vbem_iterate(0);

    compute_pi();
    compute_complete_prob();
    compute_incomplete_prob();
    ic[0] = compute_free_energy();

    release_hparam_best();
    release_free_energy_best();
    if (mode_daem) release_log_itemp();

    print_msg("  Variational Free Energy = %.Nf\n",ic[0]);

    print_ic(fp_log);
    print_seed(fp_log);
  }

  release_ic();
  release_seed();

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  etime_em = current_sec();
  etime_lap_em = elapsed_sec();
#endif

  fp_hparam = fopen(fname_hparam,"wb");
  if (fp_hparam == NULL) die("Couldn't open %s\n", fname_hparam);
  print_hparam_upd(fp_hparam);
  fclose(fp_hparam);

  if (mode_memp) {
    print_msg("Computing membership distributions...");
    fflush(fp_msg);
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
    stime_memp = current_sec();
    stime_lap_memp = elapsed_sec();
#endif
  }

  if (mode_memp || mode_predict || mode_cluster || mode_rank || mode_dissim) {
    alloc_param();
    alloc_prob_mem();

    /* for safety we run again */
    compute_pi();
    compute_complete_prob();
    compute_incomplete_prob();
    compute_conditional_prob();
    compute_expect();
    update_param_vbem();
    compute_case_member_prob();
    compute_attr_member_prob();
  }

  if (mode_memp) {
    fp_memp = fopen(fname_memp,"wb");
    if (fp_memp == NULL) die("Couldn't open %s\n", fname_memp);
    print_case_member_prob(fp_memp);
    if (mode_show_attr_memp) print_attr_member_prob(fp_memp);
    fclose(fp_memp);
  }

  if (mode_memp) {
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
    etime_memp = current_sec();
    etime_lap_memp = elapsed_sec();
#endif
    print_msg("done\n");
  }

  if (mode_predict) {
    print_msg("Computing predictive distributions...");
    fp_predict = fopen(fname_predict,"wb");
    if (fp_predict == NULL) die("Couldn't open %s\n", fname_predict);
    print_predict_prob(fp_predict);
    fclose(fp_predict);
    print_msg("done\n");
  }

  /** clustering and evaluation **/
  if (mode_cluster) cluster_main();

  /** relevance analysis **/
  if (mode_rank) rank_main();

  /** computing distances **/
  if (mode_dissim) dissim_main();

  release_hparam();
  release_hparam_gauss();
  release_hparam_upd();
  release_pi();
  release_prob();
  release_expect();
  release_free_energy();
  if (mode_memp || mode_cluster || mode_rank || mode_dissim) {
    release_prob_mem();
    release_param();
  }

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  print_time_vb(fp_log);
#endif
}

void no_vbem_main(void)
{
  FILE *fp_memp;

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_read = current_sec();
  stime_lap_read = elapsed_sec();
#endif

  read_hparam();  /* alloc_hparam() is called */
  if (mode_cluster || mode_rank) read_freq();

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  etime_read = current_sec();
  etime_lap_read = elapsed_sec();
#endif

  set_smooth_auto();
  print_size(fp_log);

  if (mode_memp) {
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
    stime_memp = current_sec();
    stime_lap_memp = elapsed_sec();
#endif
  }

  /** computing membership probs. **/
  if (mode_memp || mode_cluster || mode_rank || mode_dissim) {
    alloc_param();
    alloc_prob_mem();
    update_param_no_vbem();
    print_msg("Computing the membership distributions...");
    fflush(fp_msg);
    compute_case_member_prob();
    compute_attr_member_prob();
    print_msg("done\n");
  }
  if (mode_memp) {
    fp_memp = fopen(fname_memp,"wb"); /* binary mode */
    if (fp_memp == NULL) die("Couldn't open %s\n", fname_memp);
    print_case_member_prob(fp_memp);
    if (mode_show_attr_memp) print_attr_member_prob(fp_memp);
    fclose(fp_memp);
  }
  if (mode_memp) {
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
    etime_memp = current_sec();
    etime_lap_memp = elapsed_sec();
#endif
  }

  /** clustering & evaluation **/
  if (mode_cluster) cluster_main();

  /** relevance analysis **/
  if (mode_rank) rank_main();

  /** computing distances **/
  if (mode_dissim) dissim_main();

  /* cleaning up arrays */
  release_hparam();
  release_hparam_gauss();
  if (mode_memp || mode_cluster || mode_rank || mode_dissim) {
    release_param();
    release_prob_mem();
  }

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  print_time_vb(fp_log);
#endif
}

void vbem_iterate(int u)
{
  int r,r_best=0,n,n0,n_resume=-1;
  int num_iterate,num_iterate_best=-1;
  unsigned int curr_seed=0;

  for (r=0; r<num_init; r++) { /* loop of restarting */
    if (!mode_set_seed) {
      curr_seed = (unsigned int)random_int();
      random_seed(curr_seed);
    }

    initialize_hparam();

    n0 = (max_iterate > 0) ? max_iterate : MAX_MAX_ITERATE;
    if (num_init > 1) {
      n =
        (max_init_iterate > 0 && max_init_iterate < n0) ?
        max_init_iterate : n0;
    }
    else n = n0;
    n_resume = n0;

    num_iterate = vbem_iterate_body(r,0,n,itemp_init);

    log_seed[u][r] = (!mode_set_seed) ? curr_seed : 0;
    log_num_iterate[u][r] = num_iterate;
    log_final_free_energy[u][r] = final_free_energy;

    if (r == 0 || final_free_energy > final_free_energy_best) {
      r_best = r;
      final_free_energy_best = final_free_energy;
      num_iterate_best = num_iterate;
      save_free_energy_best(num_iterate_best);
      if (mode_daem) save_log_itemp_best(num_iterate_best);
      save_hparam_best();
      itemp_best = itemp;
    }
  }

  log_r_best[u] = r_best;

  final_free_energy = final_free_energy_best;
  restore_free_energy_best(num_iterate_best);
  if (mode_daem) restore_log_itemp_best(num_iterate_best);
  restore_hparam_best();

  num_iterate = num_iterate_best;

  if (num_init > 1) {
    print_msg("  <<Resumed best hyperparameter set #%d>>\n",r_best);
    num_iterate = vbem_iterate_body(r_best,num_iterate,n_resume,itemp_best);
  }

  print_free_energy(fp_log,num_class,num_iterate);
}

int vbem_iterate_body(int r, int m, int n, FLOAT_T itemp_init0)
{
  int mm;

  itemp = (mode_daem) ? itemp_init0 : 1.0;

  compute_pi();
  if (mode_daem && itemp < 1.0)
    compute_complete_prob_daem();
  else
    compute_complete_prob();
  compute_incomplete_prob();

  free_energy[m] =
    (mode_daem && itemp < 1.0) ?
    compute_free_energy_daem() : compute_free_energy();
  if (mode_daem) log_itemp[m] = itemp;

  if (num_init > 1)
    print_msg("  [%d] #iters ",r);
  else
    print_msg("  #iters ");

  while (1) { /* Loop for inverse temperatures */

    print_daem_message(m,itemp);
    mm = 0;

    while (1) {
      print_dot_message(m);

      compute_conditional_prob();
      compute_expect();
      update_hparam();

      m++;
      mm++;

      compute_pi();
      if (mode_daem && itemp < 1.0)
        compute_complete_prob_daem();
      else
        compute_complete_prob();
      compute_incomplete_prob();

      if (m >= free_energy_size) expand_free_energy_all();
      free_energy[m] =
        (mode_daem && itemp < 1.0) ?
        compute_free_energy_daem() : compute_free_energy();
      if (mode_daem) log_itemp[m] = itemp;

      if (is_decreased(r,m,n)) break;

      if (mode_daem && is_converged_daem(r,m,mm,n,itemp)) {
        break;
      }
      else if (!mode_daem && is_converged(r,m,n)) {
        break;
      }
    }

    if (!mode_daem && m >= n) break;

    if (itemp < 1.0) itemp *= itemp_rate;
    else break;
    if (itemp > 1.0) itemp = 1.0;
  }

  final_free_energy = free_energy[m];

  return m;
}

/* [Note] pi_c and pi_w are stored in log-scale */
void compute_pi(void)
{
  int k,v,i;
  FLOAT_T sum,psi;

  sum = 0.0;
#pragma omp parallel for reduction(+: sum)
  for (k=0; k<num_class; k++)
    sum += hparam_c_upd[k];
  psi = DIGAMMA(sum);

#pragma omp parallel for
  for (k=0; k<num_class; k++)
    pi_c[k] = DIGAMMA(hparam_c_upd[k]) - psi;

  if (mode_nominal) {
#pragma omp parallel for private(v,i,sum,psi)
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
        sum = 0.0;
        for (i=0; i<num_word[v]; i++)
          sum += hparam_w_upd[k][v][i];
        psi = DIGAMMA(sum);

        for (i=0; i<num_word[v]; i++)
          pi_w[k][v][i] = DIGAMMA(hparam_w_upd[k][v][i]) - psi;
      }
    }
  }

  if (mode_verbose) print_pi(fp_log);
}

void compute_complete_prob(void)
{
  int k,e,v,u,n,j;
  FLOAT_T sum,alpha_w0,l;
  FLOAT_T h_sdet_upd,e_c_k;
  MATRIX h_s_upd,h_p_upd,h_s_upd_obs,h_s_upd_obs_inv;
  VECTOR h_u_upd,h_u_upd_obs,x,x_obs;
  FLOAT_T xval;
  FLOAT_T h_sdet_upd_obs;
  int n_miss,n_obs,*obs;

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      prob_comp[k][e] = pi_c[k];

  if (mode_nominal) {
#pragma omp parallel for private(e,v)
    for (k=0; k<num_class; k++) {
      for (e=0; e<num_entry; e++) {
        if (total_num_miss_w > 0 && freq_list[e].num_miss_w > 0) {
          for (v=0; v<tuple_size_d; v++) {
            if (!freq_list[e].miss_w[v])
              prob_comp[k][e] += pi_w[k][v][freq_list[e].id[v]];
          }
        }
        else {
          for (v=0; v<tuple_size_d; v++)
            prob_comp[k][e] += pi_w[k][v][freq_list[e].id[v]];
        }
      }
    }
  }

  if (mode_gauss) {
#pragma omp parallel for private(e_c_k,e,u,n,alpha_w0,h_u_upd,h_s_upd,h_p_upd,h_sdet_upd,x,sum,j,l,xval,n_miss,n_obs,obs,x_obs,h_u_upd_obs,h_s_upd_obs,h_s_upd_obs_inv,h_sdet_upd_obs)
    for (k=0; k<num_class; k++) {
      e_c_k = expect_c[k];

      for (u=0; u<gauss_size; u++)
        for (e=0; e<num_entry; e++)
          prob_comp[k][e] -= 0.5 * gauss_attr_size[u] * LOG(2 * M_PI);

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        alpha_w0 = get_alpha_w(n);

        h_u_upd = hparam_u_upd[k][u];
        h_p_upd = hparam_p_upd[k][u];
        h_s_upd = hparam_s_upd[k][u];
        h_sdet_upd = hparam_sdet_upd[k][u];

        if (n == 1) { /* univariate */
          l = 0.5 * (DIGAMMA(0.5 * (alpha_w0 + e_c_k))
                     - LOG(h_sdet_upd) + LOG(2) - 1.0 / (tau + e_c_k));

          for (e=0; e<num_entry; e++) {
            n_miss = freq_list[e].num_miss_g[u];

            if (n_miss == 0) {  /* fully observed */
              prob_comp[k][e] += l;

              xval = freq_list[e].gval[u][0];
              prob_comp[k][e] -=
                0.5 * h_p_upd[0][0] * (alpha_w0 + e_c_k) *
                (xval - h_u_upd[0]) * (xval - h_u_upd[0]);
            }
            else {             /* all missing */
              prob_comp[k][e] +=
                l - 0.5 * (LOG(alpha_w0 + e_c_k) - LOG(2 * M_PI * h_sdet_upd));
            }
          }
        }
        else { /* multivariate */
          sum = 0.0;
          for (j=1; j<=n; j++) /* [Note] j should be 1...n by definition */
            sum += DIGAMMA(0.5 * (alpha_w0 + e_c_k + 1.0 - j));

          l = 0.5 * (sum - LOG(h_sdet_upd) + n * LOG(2) - n / (tau + e_c_k));

          for (e=0; e<num_entry; e++) {
            n_miss = freq_list[e].num_miss_g[u];

            if (n_miss == 0) {  /* fully observed */
              x = alloc_vector(n);

              diff_vector(freq_list[e].gval[u],h_u_upd,x);
              prob_comp[k][e] +=
                l - 0.5 * (alpha_w0 + e_c_k) * qform(h_p_upd,x);

              release_vector(x);
            }
            else if (0 < n_miss && n_miss < n) { /* partially observed */
              n_obs = (n - n_miss);
              obs = freq_list[e].obs_g[u];

              x_obs = alloc_vector(n_obs);
              h_u_upd_obs = alloc_vector(n_obs);
              h_s_upd_obs = alloc_matrix(n_obs,n_obs);
              h_s_upd_obs_inv = alloc_matrix(n_obs,n_obs);

              proj_vector(freq_list[e].gval[u],obs,x_obs);
              proj_vector(h_u_upd,obs,h_u_upd_obs);
              proj_matrix(h_s_upd,obs,obs,h_s_upd_obs);
              h_sdet_upd_obs = INVERT(h_s_upd_obs,h_s_upd_obs_inv);
              if (err_inv)
                DIE_NUMERIC("Failed to invert the covariances of the"
                            " observed part\n");

              diff_vector(x_obs,h_u_upd_obs,NULL);

              prob_comp[k][e] +=
                l - 0.5 * (alpha_w0 + e_c_k) * qform(h_s_upd_obs_inv,x_obs);
              prob_comp[k][e] -= 0.5 * (n - n_obs) * LOG(alpha_w0 + e_c_k);
              prob_comp[k][e] +=
                0.5 * (LOG(h_sdet_upd) - LOG(h_sdet_upd_obs) + (n - n_obs) * LOG(2 * M_PI));

              release_vector(x_obs);
              release_vector(h_u_upd_obs);
              release_matrix(h_s_upd_obs);
              release_matrix(h_s_upd_obs_inv);
            }
            else { /* n_miss == n, i.e. all missing */
              prob_comp[k][e] +=
                l - 0.5 * (n * LOG(alpha_w0 + e_c_k) - n * LOG(2 * M_PI) - LOG(h_sdet_upd));
            }
          }
        }
      }
    }
  }

  if (mode_verbose) print_complete_prob(fp_log);
}

void compute_complete_prob_daem(void)
{
  int k,e,v,u,n,j,n_obs,n_miss;
  int *obs;
  FLOAT_T e_c_k,alpha_w1,sum,h_sdet_upd,h_sdet_upd_obs,l,xval;
  VECTOR h_u_upd,h_u_upd_obs,x,x_obs;
  MATRIX h_s_upd,h_s_upd_obs,h_s_upd_obs_inv,h_p_upd;

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      prob_comp[k][e] = itemp * pi_c[k];

  if (mode_nominal) {
#pragma omp parallel for private(e,v)
    for (k=0; k<num_class; k++) {
      for (e=0; e<num_entry; e++) {
        if (total_num_miss_w > 0 && freq_list[e].num_miss_w > 0) {
          for (v=0; v<tuple_size_d; v++) {
            if (!freq_list[e].miss_w[v])
              prob_comp[k][e] += itemp * pi_w[k][v][freq_list[e].id[v]];
          }
        }
        else {
          for (v=0; v<tuple_size_d; v++)
            prob_comp[k][e] += itemp * pi_w[k][v][freq_list[e].id[v]];
        }
      }
    }
  }

  if (mode_gauss) {
#pragma omp parallel for private(e_c_k,e,u,n,alpha_w1,h_u_upd,h_s_upd,h_p_upd,h_sdet_upd,x,sum,j,l,xval,n_miss,n_obs,obs,x_obs,h_u_upd_obs,h_s_upd_obs,h_s_upd_obs_inv,h_sdet_upd_obs)
    for (k=0; k<num_class; k++) {
      e_c_k = expect_c[k];

      for (u=0; u<gauss_size; u++)
        for (e=0; e<num_entry; e++)
          prob_comp[k][e] -= 0.5 * itemp * gauss_attr_size[u] * LOG(2 * M_PI);

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        alpha_w1 = itemp * (get_alpha_w(n) + e_c_k - n) + n;

        h_u_upd = hparam_u_upd[k][u];
        h_s_upd = hparam_s_upd[k][u];
        h_p_upd = hparam_p_upd[k][u];
        h_sdet_upd = hparam_sdet_upd[k][u];

        if (n == 1) { /* univariate */
          l = 0.5 * (DIGAMMA(0.5 * alpha_w1) - LOG(0.5 * itemp * h_sdet_upd));
          l -= 0.5 / (itemp * (tau + e_c_k));

          for (e=0; e<num_entry; e++) {
            n_miss = freq_list[e].num_miss_g[u];

            if (n_miss == 0) {  /* fully observed */
              xval = freq_list[e].gval[u][0];
              prob_comp[k][e] +=
                itemp * (l - 0.5 * h_p_upd[0][0] * (alpha_w1 / itemp) *
                         (xval - h_u_upd[0]) * (xval - h_u_upd[0]));
            }
            else { /* all missing */
              prob_comp[k][e] +=
                itemp * (l - 0.5 * (LOG(alpha_w1 / itemp)
                                    - LOG(2 * M_PI * h_sdet_upd)));
            }
          }
        }
        else { /* multivariate */
          sum = 0.0;
          for (j=1; j<=n; j++) /* [Note] j should be 1...n by definition */
            sum += DIGAMMA(0.5 * (alpha_w1 + 1.0 - j));

          l = 0.5 * (sum - LOG(h_sdet_upd) + n * (LOG(2/itemp)));
          l -= 0.5 * (FLOAT_T)n / (itemp * (tau + e_c_k));

          for (e=0; e<num_entry; e++) {
            n_miss = freq_list[e].num_miss_g[u];

            if (n_miss == 0) {  /* fully observed */
              x = alloc_vector(n);

              diff_vector(freq_list[e].gval[u],h_u_upd,x);
              prob_comp[k][e] +=
                itemp * (l - 0.5 * (alpha_w1 / itemp) * qform(h_p_upd,x));

              release_vector(x);
            }
            else if (0 < n_miss && n_miss < n) { /* partially observed */
              n_obs = (n - n_miss);
              obs = freq_list[e].obs_g[u];

              x_obs = alloc_vector(n_obs);
              h_u_upd_obs = alloc_vector(n_obs);
              h_s_upd_obs = alloc_matrix(n_obs,n_obs);
              h_s_upd_obs_inv = alloc_matrix(n_obs,n_obs);

              proj_vector(freq_list[e].gval[u],obs,x_obs);
              proj_vector(h_u_upd,obs,h_u_upd_obs);
              proj_matrix(h_s_upd,obs,obs,h_s_upd_obs);
              h_sdet_upd_obs = INVERT(h_s_upd_obs,h_s_upd_obs_inv);
              if (err_inv)
                DIE_NUMERIC("Failed to invert the covariances of the"
                            " observed part\n");

              diff_vector(x_obs,h_u_upd_obs,NULL);

              prob_comp[k][e] +=
                itemp * (l - 0.5 * (alpha_w1 * qform(h_s_upd_obs_inv,x_obs) / itemp
                                    + (n - n_obs) * LOG(alpha_w1 / itemp)
                                    - LOG(h_sdet_upd) + LOG(h_sdet_upd_obs)
                                    - (n - n_obs) * LOG(2 * M_PI)));

              release_vector(x_obs);
              release_vector(h_u_upd_obs);
              release_matrix(h_s_upd_obs);
              release_matrix(h_s_upd_obs_inv);
            }
            else {
              prob_comp[k][e] +=
                itemp * (l - 0.5 * (n * LOG(alpha_w1 / itemp)
                                    - n * LOG(2 * M_PI) - LOG(h_sdet_upd)));
            }
          }
        }
      }
    }
  }

  if (mode_verbose) print_complete_prob(fp_log);
}

/* [Note] Complete and incomplete probs. are stored in log-scale */
void compute_incomplete_prob(void)
{
  int k,e;
  FLOAT_T pcomp_max,sum;

#pragma omp parallel for private(k,pcomp_max,sum)
  for (e=0; e<num_entry; e++) {
    pcomp_max = 0.0;
    for (k=0; k<num_class; k++)
      if (k == 0 || prob_comp[k][e] > pcomp_max) pcomp_max = prob_comp[k][e];

    sum = 0.0;
    for (k=0; k<num_class; k++)
      sum += EXP(prob_comp[k][e] - pcomp_max);

    prob_incomp[e] = pcomp_max + LOG(sum);
  }

  if (mode_verbose) print_incomplete_prob(fp_log);
}

void compute_conditional_prob(void)
{
  int k,e;

  /* [Note]
   * In VB with excessively many classes, conditional probabilities tend to
   * be very small, and the later computations become unstable.  So we forcedly
   * set PROB_COND_MIN instead of using such small conditional probabilities.
   */
#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++) {
    for (e=0; e<num_entry; e++) {
      prob_cond[k][e] = EXP(prob_comp[k][e] - prob_incomp[e]);
      if (prob_cond[k][e] < PROB_COND_MIN) prob_cond[k][e] = PROB_COND_MIN;
    }
  }

  if (mode_verbose) print_conditional_prob(fp_log);
}

/* [Note]
 * This function is already in a `for' loop computed in parallel by OpenMP.
 * So we don't need no `omp' pragma in this function.
 */
void compute_expect_gauss0_nodaem(int k, int e, FLOAT_T p_e)
{
  int u,a,a0,a1,n;
  int *rev_ind,*obs,*miss;
  int n_obs,n_miss,a_o,a_m,max_gauss_attr_size;
  FLOAT_T xval,alpha_w0,h_sdet_upd,e_c_k;
  VECTOR x,h_u_upd,x_o,hu_o,hu_m,z_o,z_m,hu_cond;
  MATRIX h_s_upd,hs_oo,hs_om,hs_mo,hs_mm,hs_oo_inv,z_mo,z_mm,hs_cond;
  MATRIX xx_oo,xx_om,xx_mo,xx_mm,xx;

  max_gauss_attr_size = -1;
  for (u=0; u<gauss_size; u++) {
    max_gauss_attr_size =
      (u == 0 || gauss_attr_size[u] > max_gauss_attr_size) ?
      gauss_attr_size[u] : max_gauss_attr_size;
  }
  rev_ind = (int *)xmalloc(sizeof(int)*max_gauss_attr_size);

  e_c_k = expect_c[k];

  for (u=0; u<gauss_size; u++) {
    n = gauss_attr_size[u];
    alpha_w0 = get_alpha_w(n);

    h_u_upd = hparam_u_upd[k][u];
    h_s_upd = hparam_s_upd[k][u];
    h_sdet_upd = hparam_sdet_upd[k][u];

    if (n == 1) { /* univariate */
      n_miss = freq_list[e].num_miss_g[u];

      if (n_miss == 0) { /* all observable */
        xval = freq_list[e].gval[u][0];
        expect_u[k][u][0] += p_e * xval;
        expect_s[k][u][0][0] += p_e * xval * xval;
      }
      else {            /* all missing */
        expect_u[k][u][0] += p_e * h_u_upd[0];
        expect_s[k][u][0][0] +=
          p_e * (h_u_upd[0] * h_u_upd[0] + h_s_upd[0][0] / (alpha_w0 + e_c_k));
      }
    }
    else { /* multivariate */
      n_miss = freq_list[e].num_miss_g[u];

      if (n_miss == 0) { /* fully observed */
        x = freq_list[e].gval[u];

        for (a=0; a<n; a++)
          expect_u[k][u][a] += p_e * x[a];

        for (a0=0; a0<n; a0++)
          for (a1=a0; a1<n; a1++) /* compute only the upper triangle */
            expect_s[k][u][a0][a1] += p_e * x[a0] * x[a1];
      }
      else if (0 < n_miss && n_miss < n) { /* partially observed */
        n_obs = (n - n_miss);
        obs = freq_list[e].obs_g[u];
        miss = freq_list[e].miss_g[u];

        /* compute the local indices */
        a_o = 0;
        a_m = 0;
        for (a=0; a<n; a++)
          rev_ind[a] = (obs[a]) ? a_o++ : a_m++;

        x_o = alloc_proj_vector_n(freq_list[e].gval[u],n_obs,obs);

        hu_o = alloc_proj_vector_n(h_u_upd,n_obs,obs);
        hu_m = alloc_proj_vector_n(h_u_upd,n_miss,miss);

        hs_oo = alloc_proj_matrix_n(h_s_upd,n_obs,n_obs,obs,obs);
        hs_om = alloc_proj_matrix_n(h_s_upd,n_obs,n_miss,obs,miss);
        hs_mo = alloc_proj_matrix_n(h_s_upd,n_miss,n_obs,miss,obs);
        hs_mm = alloc_proj_matrix_n(h_s_upd,n_miss,n_miss,miss,miss);

        hs_oo_inv = alloc_matrix(n_obs,n_obs);

        z_o = alloc_vector(n_obs);
        z_m = alloc_vector(n_miss);
        z_mo = alloc_matrix(n_miss,n_obs);
        z_mm = alloc_matrix(n_miss,n_miss);

        hu_cond = alloc_vector(n_miss);
        hs_cond = alloc_matrix(n_miss,n_miss);

        INVERT(hs_oo,hs_oo_inv);
        if (err_inv)
          DIE_NUMERIC("Failed to invert the covariances of the"
                      " observed part\n");
        prod_matrix(hs_mo,hs_oo_inv,z_mo);

        diff_vector(x_o,hu_o,z_o);
        prod_matrix_vector(z_mo,z_o,z_m);
        sum_vector(hu_m,z_m,hu_cond);

        prod_matrix(z_mo,hs_om,z_mm);
        diff_matrix(hs_mm,z_mm,hs_cond);
        div_const_matrix(hs_cond,(alpha_w0 + e_c_k));

        for (a=0; a<n; a++)
          expect_u[k][u][a] +=
            p_e * ((obs[a]) ? x_o[rev_ind[a]] : hu_cond[rev_ind[a]]);

        xx_oo = alloc_vprod_matrix(x_o,x_o);
        xx_om = alloc_vprod_matrix(x_o,hu_cond);
        xx_mo = alloc_vprod_matrix(hu_cond,x_o);
        xx_mm = alloc_vprod_matrix(hu_cond,hu_cond);
        sum_matrix(xx_mm,hs_cond,NULL);

        for (a0=0; a0<n; a0++) {
          for (a1=a0; a1<n; a1++) {
            if (obs[a0] && obs[a1])
              xx = xx_oo;
            else if (obs[a0] && miss[a1])
              xx = xx_om;
            else if (miss[a0] && obs[a1])
              xx = xx_mo;
            else
              xx = xx_mm;

            expect_s[k][u][a0][a1] += p_e * xx[rev_ind[a0]][rev_ind[a1]];
          }
        }

        release_vector(x_o);

        release_vector(hu_o);
        release_vector(hu_m);
        release_matrix(hs_oo);
        release_matrix(hs_om);
        release_matrix(hs_mo);
        release_matrix(hs_mm);
        release_matrix(hs_oo_inv);

        release_vector(z_o);
        release_vector(z_m);
        release_matrix(z_mo);
        release_matrix(z_mm);

        release_vector(hu_cond);
        release_matrix(hs_cond);

        release_matrix(xx_oo);
        release_matrix(xx_om);
        release_matrix(xx_mo);
        release_matrix(xx_mm);
      }
      else { /* n_miss == n, i.e. all missing */
        for (a=0; a<n; a++)
          expect_u[k][u][a] += p_e * h_u_upd[a];

        for (a0=0; a0<n; a0++)
          for (a1=a0; a1<n; a1++)
            expect_s[k][u][a0][a1] +=
              p_e * (h_u_upd[a0] * h_u_upd[a1] + h_s_upd[a0][a1] / (alpha_w0 + e_c_k));
      }
    }
  }

  xfree(rev_ind);
}

/* [Note]
 * This function is already in a `for' loop computed in parallel by OpenMP.
 * So we don't need no `omp' pragma in this function.
 */
void compute_expect_gauss0_daem(int k, int e, FLOAT_T p_e)
{
  int u,a,a0,a1,n;
  int *rev_ind,*obs,*miss;
  int n_obs,n_miss,a_o,a_m,max_gauss_attr_size;
  FLOAT_T xval,alpha_w0,alpha_w1,h_sdet_upd,e_c_k;
  VECTOR x,h_u_upd,x_o,hu_o,hu_m,z_o,z_m,hu_cond;
  MATRIX h_s_upd,hs_oo,hs_om,hs_mo,hs_mm,hs_oo_inv,z_mo,z_mm,hs_cond;
  MATRIX xx_oo,xx_om,xx_mo,xx_mm,xx;

  max_gauss_attr_size = -1;
  for (u=0; u<gauss_size; u++) {
    max_gauss_attr_size =
      (u == 0 || gauss_attr_size[u] > max_gauss_attr_size) ?
      gauss_attr_size[u] : max_gauss_attr_size;
  }
  rev_ind = (int *)xmalloc(sizeof(int)*max_gauss_attr_size);

  e_c_k = expect_c[k];

  for (u=0; u<gauss_size; u++) {
    n = gauss_attr_size[u];
    alpha_w0 = get_alpha_w(n);
    alpha_w1 = itemp * (alpha_w0 + e_c_k - n) + n;

    h_u_upd = hparam_u_upd[k][u];
    h_s_upd = hparam_s_upd[k][u];
    h_sdet_upd = hparam_sdet_upd[k][u];

    if (n == 1) { /* univariate */
      n_miss = freq_list[e].num_miss_g[u];

      if (n_miss == 0) { /* all observable */
        xval = freq_list[e].gval[u][0];
        expect_u[k][u][0] += p_e * xval;
        expect_s[k][u][0][0] += p_e * xval * xval;
      }
      else {            /* all missing */
        expect_u[k][u][0] += p_e * h_u_upd[0];
        expect_s[k][u][0][0] +=
          p_e * (h_u_upd[0] * h_u_upd[0] + itemp * h_s_upd[0][0] / alpha_w1);
      }
    }
    else { /* multivariate */
      n_miss = freq_list[e].num_miss_g[u];

      if (n_miss == 0) { /* fully observed */
        x = freq_list[e].gval[u];

        for (a=0; a<n; a++)
          expect_u[k][u][a] += p_e * x[a];

        for (a0=0; a0<n; a0++)
          for (a1=a0; a1<n; a1++) /* compute only the upper triangle */
            expect_s[k][u][a0][a1] += p_e * x[a0] * x[a1];
      }
      else if (0 < n_miss && n_miss < n) { /* partially observed */
        n_obs = (n - n_miss);
        obs = freq_list[e].obs_g[u];
        miss = freq_list[e].miss_g[u];

        /* compute the local indices */
        a_o = 0;
        a_m = 0;
        for (a=0; a<n; a++)
          rev_ind[a] = (obs[a]) ? a_o++ : a_m++;

        x_o = alloc_proj_vector_n(freq_list[e].gval[u],n_obs,obs);

        hu_o = alloc_proj_vector_n(h_u_upd,n_obs,obs);
        hu_m = alloc_proj_vector_n(h_u_upd,n_miss,miss);

        hs_oo = alloc_proj_matrix_n(h_s_upd,n_obs,n_obs,obs,obs);
        hs_om = alloc_proj_matrix_n(h_s_upd,n_obs,n_miss,obs,miss);
        hs_mo = alloc_proj_matrix_n(h_s_upd,n_miss,n_obs,miss,obs);
        hs_mm = alloc_proj_matrix_n(h_s_upd,n_miss,n_miss,miss,miss);

        hs_oo_inv = alloc_matrix(n_obs,n_obs);

        z_o = alloc_vector(n_obs);
        z_m = alloc_vector(n_miss);
        z_mo = alloc_matrix(n_miss,n_obs);
        z_mm = alloc_matrix(n_miss,n_miss);

        hu_cond = alloc_vector(n_miss);
        hs_cond = alloc_matrix(n_miss,n_miss);

        INVERT(hs_oo,hs_oo_inv);
        if (err_inv)
          DIE_NUMERIC("Failed to invert the covariances of"
                      " the observed part\n");
        prod_matrix(hs_mo,hs_oo_inv,z_mo);

        diff_vector(x_o,hu_o,z_o);
        prod_matrix_vector(z_mo,z_o,z_m);
        sum_vector(hu_m,z_m,hu_cond);

        prod_matrix(z_mo,hs_om,z_mm);
        diff_matrix(hs_mm,z_mm,hs_cond);
        mul_const_matrix(hs_cond,(itemp / alpha_w1));

        for (a=0; a<n; a++)
          expect_u[k][u][a] +=
            p_e * ((obs[a]) ? x_o[rev_ind[a]] : hu_cond[rev_ind[a]]);

        xx_oo = alloc_vprod_matrix(x_o,x_o);
        xx_om = alloc_vprod_matrix(x_o,hu_cond);
        xx_mo = alloc_vprod_matrix(hu_cond,x_o);
        xx_mm = alloc_vprod_matrix(hu_cond,hu_cond);
        sum_matrix(xx_mm,hs_cond,NULL);

        for (a0=0; a0<n; a0++) {
          for (a1=a0; a1<n; a1++) {
            if (obs[a0] && obs[a1])
              xx = xx_oo;
            else if (obs[a0] && miss[a1])
              xx = xx_om;
            else if (miss[a0] && obs[a1])
              xx = xx_mo;
            else
              xx = xx_mm;

            expect_s[k][u][a0][a1] += p_e * xx[rev_ind[a0]][rev_ind[a1]];
          }
        }

        release_vector(x_o);

        release_vector(hu_o);
        release_vector(hu_m);
        release_matrix(hs_oo);
        release_matrix(hs_om);
        release_matrix(hs_mo);
        release_matrix(hs_mm);
        release_matrix(hs_oo_inv);

        release_vector(z_o);
        release_vector(z_m);
        release_matrix(z_mo);
        release_matrix(z_mm);

        release_vector(hu_cond);
        release_matrix(hs_cond);

        release_matrix(xx_oo);
        release_matrix(xx_om);
        release_matrix(xx_mo);
        release_matrix(xx_mm);
      }
      else { /* n_miss == n, i.e. all missing */
        for (a=0; a<n; a++)
          expect_u[k][u][a] += p_e * h_u_upd[a];

        for (a0=0; a0<n; a0++)
          for (a1=a0; a1<n; a1++)
            expect_s[k][u][a0][a1] +=
              p_e * (h_u_upd[a0] * h_u_upd[a1] + itemp * h_s_upd[a0][a1] / alpha_w1);
      }
    }
  }

  xfree(rev_ind);
}

void compute_expect_gauss0(int k, int e, FLOAT_T p_e)
{
  if (mode_daem) compute_expect_gauss0_daem(k,e,p_e);
  else compute_expect_gauss0_nodaem(k,e,p_e);
}

void compute_expect_gauss(void)
{
  int k,e,u;

#pragma omp parallel for private(u)
  for (k=0; k<num_class; k++) {
    for (u=0; u<gauss_size; u++) {
      clear_vector(expect_u[k][u]);
      clear_matrix(expect_s[k][u]);
    }
  }

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      compute_expect_gauss0(k,e,(freq_list[e].freq * prob_cond[k][e]));

#pragma omp parallel for private(u)
  for (k=0; k<num_class; k++)
    for (u=0; u<gauss_size; u++)
      div_const_vector(expect_u[k][u],expect_c[k]);
}

void compute_expect(void)
{
  int k,i,v,e;
  FLOAT_T tmp_e;

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++) {
    expect_c[k] = 0.0;
    for (e=0; e<num_entry; e++)
      expect_c[k] += freq_list[e].freq * prob_cond[k][e];
  }

  if (mode_nominal) {
#pragma omp parallel for private(i,v)
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          expect_w[k][v][i] = 0.0;
    }

#pragma omp parallel for private(e,v,tmp_e)
    for (k=0; k<num_class; k++) {
      for (e=0; e<num_entry; e++) {
        tmp_e = freq_list[e].freq * prob_cond[k][e];
        if (total_num_miss_w > 0 && freq_list[e].num_miss_w > 0) {
          for (v=0; v<tuple_size_d; v++) {
            if (!freq_list[e].miss_w[v])
              expect_w[k][v][freq_list[e].id[v]] += tmp_e;
          }
        }
        else {
          for (v=0; v<tuple_size_d; v++)
            expect_w[k][v][freq_list[e].id[v]] += tmp_e;
        }
      }
    }
  }

  if (mode_gauss) compute_expect_gauss();

  if (mode_verbose) print_expectation(fp_log);
}

FLOAT_T compute_free_energy(void)
{
  int k,v,i,e,a,u,n;
  FLOAT_T fe;
  FLOAT_T x_u,x_s,x_u_upd,x_s_upd,x_sdet_upd,x_p_upd;
  FLOAT_T sum_c,sum_w,l,e_c_k,alpha_w0,sum_a;
  MATRIX m0,m1,m2;
  VECTOR x;

  fe = 0.0;

#pragma omp parallel for reduction(+: fe)
  for (e=0; e<num_entry; e++)
    fe += freq_list[e].freq * prob_incomp[e];

  sum_c = 0.0;
  for (k=0; k<num_class; k++)
    sum_c += hparam_c[k];

  fe += LGAMMA(sum_c);

  sum_c = 0.0;
  for (k=0; k<num_class; k++)
    sum_c += hparam_c_upd[k];

  fe -= LGAMMA(sum_c);

  for (k=0; k<num_class; k++)
    fe += LGAMMA(hparam_c_upd[k]) - LGAMMA(hparam_c[k]);

#pragma omp parallel for reduction(+: fe)
  for (k=0; k<num_class; k++)
    fe -= ((hparam_c_upd[k] - 1) - (hparam_c[k] - 1)) * pi_c[k];

  if (mode_nominal) {
#pragma omp parallel for private(v,i,sum_w) reduction(+: fe)
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
        sum_w = 0.0;
        for (i=0; i<num_word[v]; i++)
          sum_w += hparam_w[k][v][i];
        fe += LGAMMA(sum_w);

        sum_w = 0.0;
        for (i=0; i<num_word[v]; i++)
          sum_w += hparam_w_upd[k][v][i];
        fe -= LGAMMA(sum_w);

        for (i=0; i<num_word[v]; i++)
          fe += LGAMMA(hparam_w_upd[k][v][i]) - LGAMMA(hparam_w[k][v][i]);
      }
    }

#pragma omp parallel for private(v,i) reduction(+: fe)
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          fe -=
            ((hparam_w_upd[k][v][i] - 1) - (hparam_w[k][v][i] - 1)) *
            pi_w[k][v][i];
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      e_c_k = expect_c[k];

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        alpha_w0 = get_alpha_w(n);

        if (n > 1) { /* multivariate */
          l = 0.5 * n * (LOG(tau) - LOG(tau + e_c_k));
          l += lwishart_coeff(n,alpha_w0);
          l -= lwishart_coeff(n,(alpha_w0 + e_c_k));
          l += 0.5 * alpha_w0 * LOG(hparam_sdet[k][u]);
          l -= 0.5 * (alpha_w0 + e_c_k) * LOG(hparam_sdet_upd[k][u]);

          l += 0.5 * n * (e_c_k / (tau + e_c_k));

          sum_a = 0.0;
          for (a=1; a<=n; a++) /* [Note] a should be 1...n by definition */
            sum_a += DIGAMMA(0.5 * (alpha_w0 + e_c_k + 1.0 - a));

          l -= 0.5 * e_c_k * (n * LOG(2) - LOG(hparam_sdet_upd[k][u]) + sum_a);

          m0 = alloc_matrix(n,n);
          m1 = alloc_matrix(n,n);
          m2 = alloc_matrix(n,n);
          x = alloc_vector(n);

          diff_matrix(hparam_s_upd[k][u],hparam_s[k][u],m0);
          diff_vector(hparam_u_upd[k][u],hparam_u[k][u],x);

          set_vprod_matrix(m1,x,x);
          mul_const_matrix(m1,tau);
          diff_matrix(m0,m1,NULL);

          prod_matrix(m0,hparam_p_upd[k][u],m2);

          l += 0.5 * (alpha_w0 + e_c_k) * trace(m2);

          fe += l;

          release_matrix(m0);
          release_matrix(m1);
          release_matrix(m2);
          release_vector(x);
        }
        else { /* univariate */
          x_u = hparam_u[k][u][0];
          x_s = hparam_s[k][u][0][0];
          x_u_upd = hparam_u_upd[k][u][0];
          x_s_upd = hparam_s_upd[k][u][0][0];
          x_p_upd = hparam_p_upd[k][u][0][0];
          x_sdet_upd = hparam_sdet_upd[k][u];

          l = 0.5 * (LOG(tau) - LOG(tau + e_c_k));
          l += lgamma_d_coeff(0.5 * alpha_w0);
          l -= lgamma_d_coeff(0.5 * (alpha_w0 + e_c_k));
          l += 0.5 * alpha_w0 * LOG(x_s);
          l -= 0.5 * (alpha_w0 + e_c_k) * LOG(0.5 * x_s_upd);

          l += 0.5 * (e_c_k / (tau + e_c_k));

          l -= 0.5 * e_c_k *
               (LOG(2) - LOG(x_sdet_upd) + DIGAMMA(0.5 * (alpha_w0 + e_c_k)));

          l +=
            0.5 * (alpha_w0 + e_c_k) * x_p_upd *
            (x_s_upd - 2 * x_s - tau * (x_u_upd - x_u) * (x_u_upd - x_u));

          fe += l;
        }
      }
    }
  }

  return fe;
}

FLOAT_T compute_free_energy_daem(void)
{
  int k,v,i,e,u,a,n;
  FLOAT_T fe,litemp;
  FLOAT_T x_u,x_s,x_u_upd,x_s_upd,x_sdet_upd,x_p_upd;
  FLOAT_T sum_c,sum_w,l,e_c_k,e_c_k0,alpha_w0,alpha_w1,sum_a;
  MATRIX m0,m1,m2;
  VECTOR x;

  fe = 0.0;
  litemp = LOG(itemp);

#pragma omp parallel for reduction(+: fe)
  for (e=0; e<num_entry; e++)
    fe += freq_list[e].freq * prob_incomp[e];
  fe /= itemp;

  sum_c = 0.0;
  for (k=0; k<num_class; k++)
    sum_c += hparam_c[k];
  fe += LGAMMA(sum_c);

  sum_c = 0.0;
  for (k=0; k<num_class; k++)
    sum_c += hparam_c_upd[k];
  fe -= LGAMMA(sum_c)/itemp;

  for (k=0; k<num_class; k++)
    fe += LGAMMA(hparam_c_upd[k])/itemp - LGAMMA(hparam_c[k]);

  if (mode_nominal) {
#pragma omp parallel for private(v,i,sum_w) reduction(+: fe)
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
        sum_w = 0.0;
        for (i=0; i<num_word[v]; i++)
          sum_w += hparam_w[k][v][i];
        fe += LGAMMA(sum_w);

        sum_w = 0.0;
        for (i=0; i<num_word[v]; i++)
          sum_w += hparam_w_upd[k][v][i];
        fe -= LGAMMA(sum_w)/itemp;

        for (i=0; i<num_word[v]; i++)
          fe += LGAMMA(hparam_w_upd[k][v][i])/itemp - LGAMMA(hparam_w[k][v][i]);
      }
    }
  }

  for (k=0; k<num_class; k++)
    fe -= ((hparam_c_upd[k] - 1)/itemp - (hparam_c[k] - 1)) * pi_c[k];

  if (mode_nominal) {
#pragma omp parallel for private(v,i) reduction(+: fe)
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          fe -=
            ((hparam_w_upd[k][v][i] - 1)/itemp - (hparam_w[k][v][i] - 1)) *
            pi_w[k][v][i];
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      e_c_k = expect_c[k];

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        alpha_w0 = get_alpha_w(n);
        alpha_w1 = itemp * (alpha_w0 + e_c_k - n) + n;

        if (n > 1) { /* multivariate */
          l = 0.5 * n * (LOG(tau) - LOG(tau + e_c_k)/itemp
                         - litemp/itemp + (1/itemp - 1) * LOG(2 * M_PI));
          l += lwishart_coeff(n,alpha_w0);
          l -= lwishart_coeff(n,alpha_w1)/itemp;
          l += 0.5 * alpha_w0 * LOG(hparam_sdet[k][u]);
          l -= 0.5 * (alpha_w1/itemp) *
               (LOG(hparam_sdet_upd[k][u]) + n * litemp);

          l += 0.5 * n * (e_c_k / (itemp * (tau + e_c_k)));

          sum_a = 0.0;
          for (a=1; a<=n; a++) /* [Note] a should be 1...n by definition */
            sum_a += DIGAMMA(0.5 * (alpha_w1 + 1.0 - a));

          e_c_k0 = (1/itemp - 1) +
                   (alpha_w1 - n - 1)/itemp - (alpha_w0 - n - 1);
          l -= 0.5 * e_c_k0 *
               (n * LOG(2) - n * litemp - LOG(hparam_sdet_upd[k][u]) + sum_a);

          m0 = alloc_matrix(n,n);
          m1 = alloc_matrix(n,n);
          m2 = alloc_matrix(n,n);
          x = alloc_vector(n);

          diff_matrix(hparam_s_upd[k][u],hparam_s[k][u],m0);
          diff_vector(hparam_u_upd[k][u],hparam_u[k][u],x);

          set_vprod_matrix(m1,x,x);
          mul_const_matrix(m1,tau);
          diff_matrix(m0,m1,NULL);

          prod_matrix(m0,hparam_p_upd[k][u],m2);

          l += (0.5 * alpha_w1/itemp) * trace(m2);

          fe += l;

          release_matrix(m0);
          release_matrix(m1);
          release_matrix(m2);
          release_vector(x);
        }
        else { /* univariate */
          x_u = hparam_u[k][u][0];
          x_s = hparam_s[k][u][0][0];
          x_u_upd = hparam_u_upd[k][u][0];
          x_s_upd = hparam_s_upd[k][u][0][0];
          x_p_upd = hparam_p_upd[k][u][0][0];
          x_sdet_upd = hparam_sdet_upd[k][u];

          l = 0.5 * (LOG(tau) - LOG(tau + e_c_k)/itemp - litemp/itemp + (1/itemp - 1) * LOG(2 * M_PI));
          l += lgamma_d_coeff(0.5 * alpha_w0);
          l -= lgamma_d_coeff(0.5 * alpha_w1)/itemp;
          l += 0.5 * alpha_w0 * LOG(x_s);
          l -= 0.5 * (alpha_w1 / itemp) * LOG(0.5 * itemp * x_s_upd);

          l += 0.5 * e_c_k / (itemp * (tau + e_c_k));

          e_c_k0 = (1/itemp - 1) + (alpha_w1 - 2)/itemp - (alpha_w0 - 2);
          l -= 0.5 * e_c_k0 *
               (LOG(2) - litemp - LOG(x_sdet_upd) + DIGAMMA(0.5 * alpha_w1));

          l +=
            (0.5 * alpha_w1/itemp) * x_p_upd *
            (x_s_upd - 2 * x_s - tau * (x_u_upd - x_u) * (x_u_upd - x_u));

          fe += l;
        }
      }
    }
  }

  return fe;
}

static void compute_hparam_precisions(void)
{
  int k,u;

#pragma omp parallel for private(u)
  for (k=0; k<num_class; k++) {
    for (u=0; u<gauss_size; u++) {
      hparam_sdet_upd[k][u] = INVERT(hparam_s_upd[k][u],hparam_p_upd[k][u]);
      if (err_inv)
        DIE_NUMERIC("Failed to invert the hyperparameters for covariances\n");
    }
  }
}

/* [NOTE] We should not parallelize while using the random number generator. */
void initialize_hparam(void)
{
  int k,v,i,u,a,n;
  FLOAT_T r,noise;

  for (k=0; k<num_class; k++) {
    noise = noise_ratio * random_float() / num_class;
    r = smooth_c + noise;
    hparam_c[k] = r;
  }

  if (mode_nominal) {
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=(num_word[v]-1); i>=0; i--) {
          noise = noise_ratio * random_float() / num_word[v];
          r = smooth_w[v] + noise;
          hparam_w[k][v][i] = r;
        }
  }

  if (mode_gauss) {
    initialize_hparam_gauss();

    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];

        /* We only add noises to the hyperparameters for means */
        for (a=0; a<n; a++) {
          noise =
            noise_ratio_gauss * SQRT(sample_s[u][a][a]) *
            (random_float() - 0.5);
          hparam_u[k][u][a] += noise;
        }
      }

      expect_c[k] = (FLOAT_T)total_freq / num_class;
    }
  }

#pragma omp parallel for
  for (k=0; k<num_class; k++)
    hparam_c_upd[k] = hparam_c[k];

  if (mode_nominal) {
#pragma omp parallel for private(k,i)
    for (v=0; v<tuple_size_d; v++)
      for (k=0; k<num_class; k++)
        for (i=0; i<num_word[v]; i++)
          hparam_w_upd[k][v][i] = hparam_w[k][v][i];
  }

  if (mode_gauss) {
#pragma omp parallel for private(u)
    for (k=0; k<num_class; k++)
      for (u=0; u<gauss_size; u++) {
        copy_vector(hparam_u_upd[k][u],hparam_u[k][u]);
        copy_matrix(hparam_s_upd[k][u],hparam_s[k][u]);
      }
  }

  if (mode_gauss) compute_hparam_precisions();

  if (mode_verbose) print_hparam_upd(fp_log);
}

void update_param_vbem(void)
{
  int k,v,i;
  FLOAT_T sum;

  sum = 0.0;
#pragma omp parallel for reduction(+: sum)
  for (k=0; k<num_class; k++) {
    param_c[k] = hparam_c_upd[k];
    sum += hparam_c_upd[k];
  }
#pragma omp parallel for
  for (k=0; k<num_class; k++)
    param_c[k] /= sum;

  if (mode_nominal) {
#pragma omp parallel for private(v,sum,i)
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++) {
        sum = 0.0;
        for (i=0; i<num_word[v]; i++) {
          param_w[k][v][i] = hparam_w_upd[k][v][i];
          sum += hparam_w_upd[k][v][i];
        }
        for (i=0; i<num_word[v]; i++)
          param_w[k][v][i] /= sum;
      }
  }

  if (mode_verbose) print_param(fp_log);
}

void update_param_no_vbem(void)
{
  int k,v,i;
  FLOAT_T sum;

  sum = 0.0;
#pragma omp parallel for reduction(+: sum)
  for (k=0; k<num_class; k++) {
    param_c[k] = hparam_c[k];
    sum += hparam_c[k];
  }
#pragma omp parallel for
  for (k=0; k<num_class; k++)
    param_c[k] /= sum;

  if (mode_nominal) {
#pragma omp parallel for private(v,sum,i)
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
        sum = 0.0;
        for (i=0; i<num_word[v]; i++) {
          param_w[k][v][i] = hparam_w[k][v][i];
          sum += hparam_w[k][v][i];
        }
        for (i=0; i<num_word[v]; i++)
          param_w[k][v][i] /= sum;
      }
    }
  }

  if (mode_verbose) print_param(fp_log);
}

void update_hparam(void)
{
  int k,v,i,u,a,a0,a1,n;
  FLOAT_T *e_u,**e_s,e_c_k,r;
  FLOAT_T alpha_w0;
  MATRIX h_s,h_s_upd;
  VECTOR h_u,h_u_upd;
  FLOAT_T x_eu,x_es,x_hu,x_hs;

#pragma omp parallel for
  for (k=0; k<num_class; k++)
    hparam_c_upd[k] = itemp * (hparam_c[k] + expect_c[k] - 1) + 1;

  if (mode_nominal) {
#pragma omp parallel for private(k,i)
    for (v=0; v<tuple_size_d; v++)
      for (k=0; k<num_class; k++)
        for (i=0; i<num_word[v]; i++)
          hparam_w_upd[k][v][i] =
            itemp * (hparam_w[k][v][i] + expect_w[k][v][i] - 1) + 1;
  }

  if (mode_gauss) {
#pragma omp parallel for private(e_c_k,r,u,n,alpha_w0,h_u_upd,h_s_upd,e_u,e_s,h_u,h_s,a,a0,a1,x_eu,x_es,x_hu,x_hs)
    for (k=0; k<num_class; k++) {
      e_c_k = expect_c[k];
      r = (tau * e_c_k) / (tau + e_c_k);

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        alpha_w0 = get_alpha_w(n);

        h_u_upd = hparam_u_upd[k][u];
        h_s_upd = hparam_s_upd[k][u];

        if (n > 1) { /* multivariate */
          e_u = expect_u[k][u];
          e_s = expect_s[k][u];
          h_u = hparam_u[k][u];
          h_s = hparam_s[k][u];

          for (a=0; a<n; a++)
            h_u_upd[a] = (tau * h_u[a] + e_c_k * e_u[a])/(tau + e_c_k);

          for (a0=0; a0<n; a0++) {
            for (a1=a0; a1<n; a1++) {
              h_s_upd[a0][a1] = h_s[a0][a1];
              h_s_upd[a0][a1] += e_s[a0][a1] - e_c_k * e_u[a0] * e_u[a1];
              h_s_upd[a0][a1] += r * (h_u[a0] - e_u[a0]) * (h_u[a1] - e_u[a1]);
            }
          }
          copy_upper_to_lower(h_s_upd);
        }
        else { /* univariate */
          x_eu = expect_u[k][u][0];
          x_es = expect_s[k][u][0][0];
          x_hu = hparam_u[k][u][0];
          x_hs = hparam_s[k][u][0][0];

          h_u_upd[0] = (tau * x_hu + e_c_k * x_eu)/(tau + e_c_k);
          h_s_upd[0][0] =
            2 * x_hs + (x_es - e_c_k * x_eu * x_eu) +
            r * (x_hu - x_eu) * (x_hu - x_eu);
        }
      }
    }
  }

  if (mode_gauss) compute_hparam_precisions();

  if (mode_verbose) print_hparam_upd(fp_log);
}

int is_decreased(int r, int m, int n)
{
  /* This would not happen without a bug or a numeric error */
  if (mode_daem &&
      log_itemp[m] - log_itemp[m-1] < ACCEPT_SMALL_NUMBER &&
      free_energy[m] < free_energy[m-1] &&
      (free_energy[m-1] - free_energy[m]) > ACCEPT_SMALL_NUMBER) {
    print_decrease(fp_log,r,m,free_energy[m-1],free_energy[m],log_itemp[m]);
    die(conv_msg("\n(iter:%d) Variational free energy decreased! "
                 "(before: %.Nf (%.Nf) | after: %.Nf (%.Nf) | diff: %.Nf)"
                 " -- see the log file\n"),
        m,free_energy[m-1],log_itemp[m-1],
        free_energy[m],log_itemp[m],(free_energy[m-1] - free_energy[m]));
    return 1;
  }
  else if (!mode_daem &&
           free_energy[m] < free_energy[m-1] &&
           (free_energy[m-1] - free_energy[m]) > ACCEPT_SMALL_NUMBER) {
    print_decrease(fp_log,r,m,free_energy[m-1],free_energy[m],0.0);
    die(conv_msg("\n(iter:%d) Variational free energy decreased! "
                 "(before: %.Nf | after: %.Nf | diff: %.Nf)"
                 " -- see the log file\n"),
        m,free_energy[m-1],free_energy[m],(free_energy[m-1] - free_energy[m]));
    return 1;
  }
  return 0;
}

int is_converged(int r, int m, int n)
{
  if (m >= n) {
    print_msg(" (Stopped: %d iters)\tF=%.Nf\n",m,free_energy[m]);
    return 1;
  }
  else if (free_energy[m] - free_energy[m-1] < epsilon) {
    print_msg(" (Converged: %d iters)\tF=%.Nf\n",m,free_energy[m]);
    return 1;
  }
  return 0;
}

int is_converged_daem(int r, int m, int mm, int n, FLOAT_T itemp)
{
  if (itemp < 1.0) {
    if (mm >= n) {
      return 1;
    }
    else if (log_itemp[m] - log_itemp[m-1] < ACCEPT_SMALL_NUMBER &&
             free_energy[m] - free_energy[m-1] > 0 &&
             free_energy[m] - free_energy[m-1] < epsilon) {
      return 1;
    }
  }
  else {
    if (mm >= n) {
      print_msg(" (Stopped: %d iters)\tF=%.Nf\n",m,free_energy[m]);
      return 1;
    }
    else if (log_itemp[m] - log_itemp[m-1] < ACCEPT_SMALL_NUMBER &&
             free_energy[m] - free_energy[m-1] > 0 &&
             free_energy[m] - free_energy[m-1] < epsilon) {
      print_msg(" (Converged: %d iters)\tF=%.Nf\n",m,free_energy[m]);
      return 1;
    }
  }

  return 0;
}

void save_free_energy_best(int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++)
    free_energy_best[m] = free_energy[m];
}

void restore_free_energy_best(int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++)
    free_energy[m] = free_energy_best[m];
}

void save_log_itemp_best(int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++)
    log_itemp_best[m] = log_itemp[m];
}

void restore_log_itemp_best(int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++)
    log_itemp[m] = log_itemp_best[m];
}
